import json
import os
import re
import sys
import time
from abc import ABC
from collections.abc import Iterable, Iterator
from concurrent.futures import ThreadPoolExecutor
from copy import copy
from functools import cached_property, wraps
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, Optional, Union, cast

import ijson  # type: ignore
import requests
from eth_pydantic_types import HexBytes
from eth_typing import BlockNumber, HexStr
from eth_utils import add_0x_prefix, is_hex, to_hex
from evmchains import PUBLIC_CHAIN_META, get_random_rpc
from pydantic.dataclasses import dataclass
from requests import HTTPError
from web3 import HTTPProvider, IPCProvider, Web3
from web3 import __version__ as web3_version
from web3.exceptions import ContractLogicError as Web3ContractLogicError
from web3.exceptions import (
    ExtraDataLengthError,
    MethodUnavailable,
    TimeExhausted,
    TransactionNotFound,
)

from ape.types.private_mempool import Bundle, SimulationReport

try:
    from web3.exceptions import Web3RPCError  # type: ignore
except ImportError:
    Web3RPCError = ValueError  # type: ignore

from web3.gas_strategies.rpc import rpc_gas_price_strategy
from web3.middleware.validation import MAX_EXTRADATA_LENGTH
from web3.providers import AutoProvider
from web3.providers.auto import load_provider_from_environment
from web3.types import FeeHistory, RPCEndpoint, TxParams

from ape.api.address import Address
from ape.api.providers import BlockAPI, ProviderAPI
from ape.api.transactions import ReceiptAPI, TransactionAPI
from ape.exceptions import (
    _SOURCE_TRACEBACK_ARG,
    _TRACE_ARG,
    ApeException,
    APINotImplementedError,
    BlockNotFoundError,
    ConfigError,
    ContractLogicError,
    ContractNotFoundError,
    OutOfGasError,
    ProviderError,
    ProviderNotConnectedError,
    TransactionError,
    TransactionNotFoundError,
    VirtualMachineError,
)
from ape.logging import logger, sanitize_url
from ape.types.events import ContractLog, LogFilter
from ape.types.gas import AutoGasLimit
from ape.types.trace import SourceTraceback
from ape.utils._web3_compat import ExtraDataToPOAMiddleware, WebsocketProvider
from ape.utils.basemodel import ManagerAccessMixin
from ape.utils.misc import DEFAULT_MAX_RETRIES_TX, gas_estimation_error_message, to_int
from ape.utils.rpc import request_with_retry
from ape_ethereum._print import CONSOLE_ADDRESS, console_contract
from ape_ethereum.trace import CallTrace, TraceApproach, TransactionTrace
from ape_ethereum.transactions import AccessList, AccessListTransaction, TransactionStatusEnum

if TYPE_CHECKING:
    from ethpm_types import EventABI

    from ape.api.trace import TraceAPI
    from ape.types.address import AddressType
    from ape.types.vm import BlockID, ContractCode


DEFAULT_PORT = 8545
DEFAULT_HOSTNAME = "localhost"
DEFAULT_HTTP_URI = f"http://{DEFAULT_HOSTNAME}:{DEFAULT_PORT}"
DEFAULT_SETTINGS = {"uri": DEFAULT_HTTP_URI}


def _sanitize_web3_url(msg: str) -> str:
    """Sanitize RPC URI from given log string"""

    # `auto` used by some providers to figure it out automatically
    if "URI: " not in msg or "URI: auto" in msg:
        return msg

    parts = msg.split("URI: ")
    prefix = parts[0].strip()
    rest = parts[1].split(" ")

    # * To remove the `,` from the url http://127.0.0.1:8545,
    if "," in rest[0]:
        rest[0] = rest[0].rstrip(",")
    sanitized_url = sanitize_url(rest[0])
    return f"{prefix} URI: {sanitized_url} {' '.join(rest[1:])}"


WEB3_PROVIDER_URI_ENV_VAR_NAME = "WEB3_PROVIDER_URI"


def assert_web3_provider_uri_env_var_not_set():
    """
    Environment variable $WEB3_PROVIDER_URI causes problems
    when used with Ape (ignores Ape's networks). Use
    this validator to eliminate the concern.

    Raises:
          :class:`~ape.exceptions.ProviderError`: If environment variable
            WEB3_PROVIDER_URI exists in ``os.environ``.
    """
    if WEB3_PROVIDER_URI_ENV_VAR_NAME not in os.environ:
        return

    # NOTE: This was the source of confusion for user when they noticed
    #  Ape would only connect to RPC URL set by an environment variable
    #  named $WEB3_PROVIDER_URI instead of whatever network they were telling Ape.
    raise ProviderError(
        "Ape does not support Web3.py's environment variable "
        f"${WEB3_PROVIDER_URI_ENV_VAR_NAME}. If you are using this environment "
        "variable name incidentally, please use a different name. If you are "
        "trying to set the network in Web3.py, please use Ape's `ape-config.yaml` "
        "or `--network` option instead."
    )


def _post_send_transaction(tx: TransactionAPI, receipt: ReceiptAPI):
    """Execute post-transaction ops"""

    # TODO: Optional configuration?
    if tx.receiver and Address(tx.receiver).is_contract:
        # Look for and print any contract logging
        try:
            receipt.show_debug_logs()
        except TransactionNotFound:
            # Receipt never published. Likely failed.
            pass
        except Exception as err:
            # Avoid letting debug logs causes program crashes.
            logger.debug(f"Unable to show debug logs: {err}")

    logger.info(f"Confirmed {receipt.txn_hash} (total fees paid = {receipt.total_fees_paid})")


class Web3Provider(ProviderAPI, ABC):
    """
    A base provider mixin class that uses the
    `web3.py <https://web3py.readthedocs.io/en/stable/>`__ python package.
    """

    _web3: Optional[Web3] = None
    _client_version: Optional[str] = None

    _call_trace_approach: Optional[TraceApproach] = None
    """
    Is ``None`` until known.
    NOTE: This gets set in `ape_ethereum.trace.Trace`.
    """

    _supports_debug_trace_call: Optional[bool] = None

    _transaction_trace_cache: dict[str, TransactionTrace] = {}

    def __new__(cls, *args, **kwargs):
        assert_web3_provider_uri_env_var_not_set()

        # Post-connection ops
        def post_connect_hook(connect):
            @wraps(connect)
            def connect_wrapper(self):
                connect(self)
                self._post_connect()

            return connect_wrapper

        # Patching the provider to call a post send_transaction() hook
        def post_tx_hook(send_tx):
            if getattr(send_tx, "_is_post_tx_wrapped", False):
                return send_tx

            @wraps(send_tx)
            def send_tx_wrapper(self, txn: TransactionAPI) -> ReceiptAPI:
                receipt = send_tx(self, txn)
                _post_send_transaction(txn, receipt)
                return receipt

            send_tx_wrapper._is_post_tx_wrapped = True  # type: ignore
            return send_tx_wrapper

        send_tx_wrapper = post_tx_hook(cls.send_transaction)
        connect_wrapper = post_connect_hook(cls.connect)
        cls.send_transaction = send_tx_wrapper  # type: ignore[method-assign]
        cls.connect = connect_wrapper  # type: ignore[method-assign]
        return super().__new__(cls)  # pydantic v2 doesn't want args

    def __init__(self, *args, **kwargs):
        logger.create_logger("web3.RequestManager", handlers=(_sanitize_web3_url,))
        logger.create_logger("web3.providers.HTTPProvider", handlers=(_sanitize_web3_url,))
        super().__init__(*args, **kwargs)

    @property
    def web3(self) -> Web3:
        """
        Access to the ``web3`` object as if you did ``Web3(HTTPProvider(uri))``.
        """

        if web3 := self._web3:
            return web3

        raise ProviderNotConnectedError()

    @property
    def _network_config(self) -> dict:
        config: dict = self.settings.get(self.network.ecosystem.name, None)
        if config is None:
            return {}

        return (config or {}).get(self.network.name) or {}

    def _get_configured_rpc(self, key: str, validator: Callable[[str], bool]) -> Optional[str]:
        # key = "uri", "http_uri", "ws_uri", or "ipc_path"
        settings = self.settings  # Includes self.provider_settings and top-level config.
        result = None
        rpc: str
        if rpc := settings.get(key):
            result = f"{rpc}"

        else:
            # See if it was configured for the network directly.
            config = self._network_config
            if rpc := config.get(key):
                result = f"{rpc}"

        if result:
            if validator(result):
                return result
            else:
                raise ConfigError(f"Invalid {key}: {result}")

        # Not configured by the user.
        return None

    @property
    def _configured_http_uri(self) -> Optional[str]:
        return self._get_configured_rpc("http_uri", _is_http_url)

    @property
    def _configured_ws_uri(self) -> Optional[str]:
        return self._get_configured_rpc("ws_uri", _is_ws_url)

    @property
    def _configured_ipc_path(self) -> Optional[str]:
        return self._get_configured_rpc("ipc_path", _is_ipc_path)

    @property
    def _configured_uri(self) -> Optional[str]:
        for key in ("uri", "url", "ipc_path", "http_uri", "ws_uri"):
            if rpc := self._get_configured_rpc(key, _is_uri):
                return rpc

        return None

    @property
    def _configured_rpc(self) -> Optional[str]:
        """
        First of URI, HTTP_URI, WS_URI, IPC_PATH
        found in the provider_settings or config.
        """

        # NOTE: Even though this only returns 1 value,
        #   each configured URI is passed in to web3 and
        #   will be used as each specific types of data
        #   is requested.
        if rpc := self._configured_uri:
            # The user specifically configured "uri:"
            return rpc

        elif rpc := self._configured_http_uri:
            # Use their configured HTTP URI.
            return rpc

        elif rpc := self._configured_ws_uri:
            # Use their configured WS URI.
            return rpc

        elif rpc := self._configured_ipc_path:
            return rpc

        return None

    def _get_connected_rpc(self, validator: Callable[[str], bool]) -> Optional[str]:
        """
        The connected HTTP URI. If using providers
        like `ape-node`, configure your URI and that will
        be returned here instead.
        """
        if web3 := self._web3:
            if endpoint_uri := getattr(web3.provider, "endpoint_uri", None):
                if isinstance(endpoint_uri, str) and validator(endpoint_uri):
                    return endpoint_uri

        return None

    @property
    def _connected_http_uri(self) -> Optional[str]:
        return self._get_connected_rpc(_is_http_url)

    @property
    def _connected_ws_uri(self) -> Optional[str]:
        return self._get_connected_rpc(_is_ws_url)

    @property
    def _connected_ipc_path(self) -> Optional[str]:
        return self._get_connected_rpc(_is_ipc_path)

    @property
    def _connected_uri(self) -> Optional[str]:
        return self._get_connected_rpc(_is_uri)

    @property
    def uri(self) -> str:
        if rpc := self._connected_uri:
            # The already connected RPC URI.
            return rpc

        elif rpc := self._configured_rpc:
            # Any configured rpc from settings/config.
            return rpc

        elif rpc := self._default_http_uri:
            # Default localhost RPC or random chain from `evmchains`
            # (depending on network).
            return rpc

        # NOTE: Don't use default IPC path here. IPC must be
        #   configured if it is the only RPC.

        raise ProviderError(
            f"Missing URI for network '{self.network.name}' on '{self.network.ecosystem.name}'."
        )

    @property
    def network_choice(self) -> str:
        if uri := self._configured_uri:
            # Ensure anything using the same choice uses the same RPC.
            if self.network.name == "custom":
                # Network was not really specified. Just use URI.
                return uri

            # User is using a value like `ethereum:mainnet:<uri>` or
            # configured the URI in their Ape config.
            return f"{self.network.choice}:{uri}"

        return super().network_choice

    @property
    def http_uri(self) -> Optional[str]:
        if rpc := self._connected_http_uri:
            return rpc

        elif rpc := self._configured_http_uri:
            return rpc

        elif rpc := self._configured_uri:
            if _is_http_url(rpc):
                # "uri" found in config/settings and is WS.
                return rpc

        return self._default_http_uri

    @property
    def _default_http_uri(self) -> Optional[str]:
        if self.network.is_dev:
            # Nothing is configured and we are running geth --dev.
            # Use a default localhost value.
            return DEFAULT_HTTP_URI

        elif rpc := self._get_random_rpc():
            # This works when the network is in `evmchains`.
            return rpc

        return None

    @property
    def ws_uri(self) -> Optional[str]:
        if rpc := self._connected_ws_uri:
            return rpc

        elif rpc := self._configured_ws_uri:
            # "ws_uri" found in config/settings
            return rpc

        elif rpc := self._configured_uri:
            if _is_ws_url(rpc):
                # "uri" found in config/settings and is WS.
                return rpc

        return None

    @property
    def ipc_path(self) -> Optional[Path]:
        if rpc := self._configured_ipc_path:
            # "ipc_path" found in config/settings
            return Path(rpc)

        elif rpc := self._configured_uri:
            if _is_ipc_path(rpc):
                # "uri" found in config/settings and is IPC.
                return Path(rpc)

        return None

    def _get_random_rpc(self) -> Optional[str]:
        if self.network.is_dev:
            return None

        ecosystem = self.network.ecosystem.name
        network = self.network.name

        # Use public RPC if available
        try:
            rpc = get_random_rpc(ecosystem, network)
        except KeyError:
            return None

        def rpc_available(rpc_uri: str) -> bool:
            try:
                return Web3(HTTPProvider(rpc_uri)).is_connected()

            except Exception:
                return False

        retries = 10
        while retries > 0:
            if rpc_available(rpc):
                return rpc

            rpc = get_random_rpc(ecosystem, network)
            logger.warning(f"RPC at '{rpc}' not available, retrying {retries} more times")
            retries -= 1

        return None

    @property
    def client_version(self) -> str:
        if not self._web3:
            return ""

        # NOTE: Gets reset to `None` on `connect()` and `disconnect()`.
        if self._client_version is None:
            self._client_version = self.web3.client_version

        return self._client_version

    @property
    def base_fee(self) -> int:
        latest_block_number = self._get_latest_block_rpc().get("number")
        if latest_block_number is None:
            # Possibly no blocks yet.
            logger.debug("Latest block has no number. Using base fee of '0'.")
            return 0

        try:
            fee_history = self._get_fee_history(latest_block_number)
        except Exception as exc:
            # Use the less-accurate approach (OK for testing).
            logger.debug(
                "Failed using `web3.eth.fee_history` for network "
                f"'{self.network_choice}'. Error: {exc}"
            )
            return self._get_last_base_fee()

        if "baseFeePerGas" not in fee_history or len(fee_history["baseFeePerGas"] or []) < 2:
            logger.debug("Not enough fee_history. Defaulting less-accurate approach.")
            return self._get_last_base_fee()

        pending_base_fee = fee_history["baseFeePerGas"][1]
        if pending_base_fee is None:
            # Non-EIP-1559 chains or we time-travelled pre-London fork.
            return self._get_last_base_fee()

        return to_int(pending_base_fee)

    @property
    def call_trace_approach(self) -> Optional[TraceApproach]:
        """
        The default tracing approach to use when building up a call-tree.
        By default, Ape attempts to use the faster approach. Meaning, if
        geth-call-tracer or parity are available, Ape will use one of those
        instead of building a call-trace entirely from struct-logs.
        """
        if approach := self._call_trace_approach:
            return approach

        return self.settings.get("call_trace_approach")

    def _get_fee_history(self, block_number: int) -> FeeHistory:
        try:
            return self.web3.eth.fee_history(1, BlockNumber(block_number), reward_percentiles=[])
        except (MethodUnavailable, AttributeError) as err:
            raise APINotImplementedError(str(err)) from err

    def _get_last_base_fee(self) -> int:
        base_fee = self._get_latest_block_rpc().get("baseFeePerGas", None)
        if base_fee is not None:
            return to_int(base_fee)

        raise APINotImplementedError("No base fee found in block.")

    @property
    def is_connected(self) -> bool:
        if self._web3 is None:
            return False

        return self._web3.is_connected()

    @property
    def max_gas(self) -> int:
        return int(self._get_latest_block_rpc()["gasLimit"], 16)

    @cached_property
    def supports_tracing(self) -> bool:
        try:
            # NOTE: Txn hash is purposely not a real hash.
            self.make_request("debug_traceTransaction", ["__CHECK_IF_SUPPORTS_TRACING__"])
        except NotImplementedError:
            return False

        except Exception:
            # We know tracing works because we didn't get a NotImplementedError.
            return True

        return True

    def update_settings(self, new_settings: dict):
        self.disconnect()
        self.provider_settings.update(new_settings)
        self.connect()

    def estimate_gas_cost(self, txn: TransactionAPI, block_id: Optional["BlockID"] = None) -> int:
        # NOTE: Using JSON mode since used as request data.
        txn_dict = txn.model_dump(by_alias=True, mode="json")

        # Force the use of hex values to support a wider range of nodes.
        if isinstance(txn_dict.get("type"), int):
            txn_dict["type"] = to_hex(txn_dict["type"])

        # NOTE: "auto" means to enter this method, so remove it from dict
        if "gas" in txn_dict and (
            txn_dict["gas"] == "auto" or isinstance(txn_dict["gas"], AutoGasLimit)
        ):
            txn_dict.pop("gas")
            # Also pop these, they are overridden by "auto"
            txn_dict.pop("maxFeePerGas", None)
            txn_dict.pop("maxPriorityFeePerGas", None)

        txn_params = cast(TxParams, txn_dict)
        try:
            return self.web3.eth.estimate_gas(txn_params, block_identifier=block_id)
        except (ValueError, Web3ContractLogicError) as err:
            # NOTE: Try to use debug_traceCall to obtain a trace.
            #  And the RPC can be very picky with inputs.
            tx_to_trace: dict = {}
            for key, val in txn_params.items():
                if isinstance(val, int):
                    tx_to_trace[key] = hex(val)
                else:
                    tx_to_trace[key] = val

            tx_error = self.get_virtual_machine_error(
                err,
                txn=txn,
                trace=lambda: CallTrace(tx=txn),
                set_ape_traceback=False,
            )

            # If this is the cause of a would-be revert,
            # raise ContractLogicError so that we can confirm tx-reverts.
            if isinstance(tx_error, ContractLogicError):
                raise tx_error from err

            message = gas_estimation_error_message(tx_error)
            raise TransactionError(
                message,
                base_err=tx_error,
                txn=txn,
                source_traceback=lambda: tx_error.source_traceback,
                set_ape_traceback=True,
            ) from err

    @cached_property
    def chain_id(self) -> int:
        default_chain_id = None
        if not self.network.is_adhoc and self.network.is_custom:
            # If using a configured, custom network, the chain ID is hardcoded.
            default_chain_id = self.network.chain_id

        elif not self.network.is_dev and not self.network.is_adhoc:
            # If using a configured public network, the chain ID is hardcoded.
            default_chain_id = self.network.chain_id

        try:
            if hasattr(self.web3, "eth"):
                return self._get_chain_id()

        except ProviderNotConnectedError:
            if default_chain_id is not None:
                return default_chain_id

            raise  # Original error

        except ValueError as err:
            # Possible syncing error.
            raise ProviderError(
                err.args[0].get("message")
                if all((hasattr(err, "args"), err.args, isinstance(err.args[0], dict)))
                else "Error getting chain ID."
            )

        if default_chain_id is not None:
            return default_chain_id

        raise ProviderNotConnectedError()

    @property
    def gas_price(self) -> int:
        price = self.web3.eth.generate_gas_price() or 0
        return to_int(price)

    @property
    def priority_fee(self) -> int:
        try:
            return self.web3.eth.max_priority_fee
        except MethodUnavailable as err:
            # The user likely should be using a more-catered plugin.
            raise APINotImplementedError(
                "eth_maxPriorityFeePerGas not supported in this RPC. Please specify manually."
            ) from err

    def _get_chain_id(self) -> int:
        result = self.make_request("eth_chainId", [])
        return result if isinstance(result, int) else int(result, 16)

    def get_block(self, block_id: "BlockID") -> BlockAPI:
        if isinstance(block_id, str) and block_id.isnumeric():
            block_id = int(block_id)

        try:
            block_data = dict(self.web3.eth.get_block(block_id))
        except Exception as err:
            raise BlockNotFoundError(block_id, reason=str(err)) from err

        return self.network.ecosystem.decode_block(block_data)

    def _get_latest_block(self) -> BlockAPI:
        # perf: By-pass as much as possible since this is a common action.
        data = self._get_latest_block_rpc()
        return self.network.ecosystem.decode_block(data)

    def _get_latest_block_rpc(self) -> dict:
        return self.make_request("eth_getBlockByNumber", ["latest", False])

    def get_nonce(self, address: "AddressType", block_id: Optional["BlockID"] = None) -> int:
        return self.web3.eth.get_transaction_count(address, block_identifier=block_id)

    def get_balance(self, address: "AddressType", block_id: Optional["BlockID"] = None) -> int:
        return self.web3.eth.get_balance(address, block_identifier=block_id)

    def get_code(
        self, address: "AddressType", block_id: Optional["BlockID"] = None
    ) -> "ContractCode":
        return self.web3.eth.get_code(address, block_identifier=block_id)

    def get_storage(
        self, address: "AddressType", slot: int, block_id: Optional["BlockID"] = None
    ) -> HexBytes:
        try:
            return HexBytes(self.web3.eth.get_storage_at(address, slot, block_identifier=block_id))
        except ValueError as err:
            if "RPC Endpoint has not been implemented" in str(err):
                raise APINotImplementedError(str(err)) from err

            raise  # Raise original error

    def get_transaction_trace(self, transaction_hash: str, **kwargs) -> "TraceAPI":
        if transaction_hash in self._transaction_trace_cache:
            return self._transaction_trace_cache[transaction_hash]

        if "call_trace_approach" not in kwargs:
            kwargs["call_trace_approach"] = self.call_trace_approach

        trace = TransactionTrace(transaction_hash=transaction_hash, **kwargs)
        self._transaction_trace_cache[transaction_hash] = trace
        return trace

    def send_call(
        self,
        txn: TransactionAPI,
        block_id: Optional["BlockID"] = None,
        state: Optional[dict] = None,
        **kwargs: Any,
    ) -> HexBytes:
        if block_id is not None:
            kwargs["block_identifier"] = block_id

        if state is not None:
            kwargs["state_override"] = state

        raise_on_revert = kwargs.get("raise_on_revert", txn.raise_on_revert)
        skip_trace = kwargs.pop("skip_trace", False)
        arguments = self._prepare_call(txn, **kwargs)
        if skip_trace:
            return self._eth_call(
                arguments, raise_on_revert=txn.raise_on_revert, skip_trace=skip_trace
            )

        show_gas = kwargs.pop("show_gas_report", False)
        show_trace = kwargs.pop("show_trace", False)

        if self._test_runner is not None:
            track_gas = self._test_runner.gas_tracker.enabled
            track_coverage = self._test_runner.coverage_tracker.enabled
        else:
            track_gas = False
            track_coverage = False

        needs_trace = track_gas or track_coverage or show_gas or show_trace
        if not needs_trace:
            return self._eth_call(arguments, raise_on_revert=raise_on_revert, skip_trace=skip_trace)

        # The user is requesting information related to a call's trace,
        # such as gas usage data.

        # When looking at gas, we cannot use token symbols in enrichment.
        # Else, the table is difficult to understand.
        use_symbol_for_tokens = track_gas or show_gas

        trace = CallTrace(
            tx=arguments[0],
            arguments=arguments[1:],
            use_symbol_for_tokens=use_symbol_for_tokens,
            supports_debug_trace_call=self._supports_debug_trace_call,
        )

        if track_gas and self._test_runner is not None and txn.receiver:
            self._test_runner.gas_tracker.append_gas(trace, txn.receiver)

        if track_coverage and self._test_runner is not None and txn.receiver:
            if contract_type := self.chain_manager.contracts.get(txn.receiver):
                if contract_src := self.local_project._create_contract_source(contract_type):
                    method_id = HexBytes(txn.data)
                    selector = (
                        contract_type.methods[method_id].selector
                        if method_id in contract_type.methods
                        else None
                    )
                    source_traceback = SourceTraceback.create(contract_src, trace, method_id)
                    self._test_runner.coverage_tracker.cover(
                        source_traceback, function=selector, contract=contract_type.name
                    )

        if show_gas:
            trace.show_gas_report()

        if show_trace:
            trace.show()

        return HexBytes(trace.return_value)

    def _eth_call(
        self, arguments: list, raise_on_revert: bool = True, skip_trace: bool = False
    ) -> HexBytes:
        # Force the usage of hex-type to support a wider-range of nodes.
        txn_dict = copy(arguments[0])
        if isinstance(txn_dict.get("type"), int):
            txn_dict["type"] = to_hex(txn_dict["type"])

        # Remove unnecessary values to support a wider-range of nodes.
        txn_dict.pop("chainId", None)

        arguments[0] = txn_dict
        try:
            result = self.make_request("eth_call", arguments)
        except Exception as err:
            contract_address = arguments[0].get("to")
            _lazy_call_trace = _LazyCallTrace(arguments)

            if not skip_trace:
                if address := contract_address:
                    try:
                        contract_type = self.chain_manager.contracts.get(address)
                    except RecursionError:
                        # Occurs when already in the middle of fetching this contract.
                        pass
                    else:
                        _lazy_call_trace.contract_type = contract_type

            vm_err = self.get_virtual_machine_error(
                err,
                trace=lambda: _lazy_call_trace.trace,
                contract_address=contract_address,
                source_traceback=lambda: _lazy_call_trace.source_traceback,
                set_ape_traceback=raise_on_revert,
            )
            if raise_on_revert:
                raise vm_err.with_ape_traceback() from err

            else:
                logger.error(vm_err)
                result = "0x"

        if "error" in result:
            raise ProviderError(result["error"]["message"])

        return HexBytes(result)

    def _prepare_call(self, txn: Union[dict, TransactionAPI], **kwargs) -> list:
        # NOTE: Using mode="json" because used in request data.
        txn_dict = (
            txn.model_dump(by_alias=True, mode="json") if isinstance(txn, TransactionAPI) else txn
        )
        fields_to_convert = ("data", "chainId", "value")
        for field in fields_to_convert:
            value = txn_dict.get(field)
            if value is not None and not isinstance(value, str):
                txn_dict[field] = to_hex(value)

        # Remove unneeded properties
        txn_dict.pop("gas", None)
        txn_dict.pop("gasLimit", None)
        txn_dict.pop("maxFeePerGas", None)
        txn_dict.pop("maxPriorityFeePerGas", None)
        txn_dict.pop("signature", None)

        # NOTE: Block ID is required so if given None, default to `"latest"`.
        block_identifier = kwargs.pop("block_identifier", kwargs.pop("block_id", None)) or "latest"
        if isinstance(block_identifier, int):
            block_identifier = to_hex(primitive=block_identifier)

        arguments = [txn_dict, block_identifier]
        if "state_override" in kwargs:
            arguments.append(kwargs["state_override"])

        return arguments

    def get_receipt(
        self,
        txn_hash: str,
        required_confirmations: int = 0,
        timeout: Optional[int] = None,
        **kwargs,
    ) -> ReceiptAPI:
        if required_confirmations < 0:
            raise TransactionError("Required confirmations cannot be negative.")

        timeout = (
            timeout if timeout is not None else self.provider.network.transaction_acceptance_timeout
        )
        hex_hash = HexBytes(txn_hash)

        txn: dict = {}
        if transaction := kwargs.get("transaction"):
            # perf: If called `send_transaction()`, we should already have the data!
            txn = (
                transaction
                if isinstance(transaction, dict)
                else transaction.model_dump(by_alias=True, mode="json")
            )

        private = kwargs.get("private")

        try:
            receipt_data = dict(
                self.web3.eth.wait_for_transaction_receipt(hex_hash, timeout=timeout)
            )
        except TimeExhausted as err:
            # Since private transactions can take longer,
            #  return a partial receipt instead of throwing a TimeExhausted error.
            if private:
                # Return with a partial receipt
                data = {
                    "block_number": -1,
                    "required_confirmations": required_confirmations,
                    "txn_hash": txn_hash,
                    "status": TransactionStatusEnum.NO_ERROR,
                    **txn,
                }
                receipt = self._create_receipt(**data)
                return receipt
            msg_str = str(err)
            if f"HexBytes('{txn_hash}')" in msg_str:
                msg_str = msg_str.replace(f"HexBytes('{txn_hash}')", f"'{txn_hash}'")

            raise TransactionNotFoundError(
                transaction_hash=txn_hash, error_message=msg_str
            ) from err

        ecosystem_config = self.network.ecosystem_config
        network_config: dict = ecosystem_config.get(self.network.name, {})
        max_retries = network_config.get("max_get_transaction_retries", DEFAULT_MAX_RETRIES_TX)

        if transaction:
            if "effectiveGasPrice" in receipt_data:
                receipt_data["gasPrice"] = receipt_data["effectiveGasPrice"]

        else:
            for attempt in range(max_retries):
                try:
                    txn = dict(self.web3.eth.get_transaction(HexStr(txn_hash)))
                    break

                except TransactionNotFound:
                    if attempt < max_retries - 1:
                        # Not the last attempt. Wait and then retry.
                        time.sleep(0.5)
                        continue

                    else:
                        # It was the last attempt - raise the exception as-is.
                        raise

        data = {"required_confirmations": required_confirmations, **txn, **receipt_data}
        receipt = self._create_receipt(**data)
        return receipt.await_confirmations()

    def _create_receipt(self, **kwargs) -> ReceiptAPI:
        data = {"provider": self, **kwargs}
        return self.network.ecosystem.decode_receipt(data)

    def get_transactions_by_block(self, block_id: "BlockID") -> Iterator[TransactionAPI]:
        if isinstance(block_id, str):
            block_id = HexStr(block_id)

            if block_id.isnumeric():
                block_id = add_0x_prefix(block_id)

        block = cast(dict, self.web3.eth.get_block(block_id, full_transactions=True))
        for transaction in block.get("transactions", []):
            yield self.network.ecosystem.create_transaction(**transaction)

    def get_transactions_by_account_nonce(
        self,
        account: "AddressType",
        start_nonce: int = 0,
        stop_nonce: int = -1,
    ) -> Iterator[ReceiptAPI]:
        if start_nonce > stop_nonce:
            raise ValueError("Starting nonce cannot be greater than stop nonce for search")

        if not self.network.is_local and (stop_nonce - start_nonce) > 2:
            # NOTE: RPC usage might be acceptable to find 1 or 2 transactions reasonably quickly
            logger.warning(
                "Performing this action is likely to be very slow and may "
                f"use {20 * (stop_nonce - start_nonce)} or more RPC calls. "
                "Consider installing an alternative data query provider plugin."
            )

        yield from self._find_txn_by_account_and_nonce(
            account,
            start_nonce,
            stop_nonce,
            0,  # first block
            self.chain_manager.blocks.head.number or 0,  # last block (or 0 if genesis-only chain)
        )

    def _find_txn_by_account_and_nonce(
        self,
        account: "AddressType",
        start_nonce: int,
        stop_nonce: int,
        start_block: int,
        stop_block: int,
    ) -> Iterator[ReceiptAPI]:
        # binary search between `start_block` and `stop_block` to yield txns from account,
        # ordered from `start_nonce` to `stop_nonce`

        if start_block == stop_block:
            # Honed in on one block where there's a delta in nonce, so must be the right block
            for txn in self.get_transactions_by_block(stop_block):
                assert isinstance(txn.nonce, int)  # NOTE: just satisfying mypy here
                if txn.sender == account and txn.nonce >= start_nonce:
                    yield self.get_receipt(to_hex(txn.txn_hash))

            # Nothing else to search for

        else:
            # Break up into smaller chunks
            # NOTE: biased to `stop_block`
            block_number = start_block + (stop_block - start_block) // 2 + 1
            txn_count_prev_to_block = self.web3.eth.get_transaction_count(account, block_number - 1)

            if start_nonce < txn_count_prev_to_block:
                yield from self._find_txn_by_account_and_nonce(
                    account,
                    start_nonce,
                    min(txn_count_prev_to_block - 1, stop_nonce),  # NOTE: In case >1 txn in block
                    start_block,
                    block_number - 1,
                )

            if txn_count_prev_to_block <= stop_nonce:
                yield from self._find_txn_by_account_and_nonce(
                    account,
                    max(start_nonce, txn_count_prev_to_block),  # NOTE: In case >1 txn in block
                    stop_nonce,
                    block_number,
                    stop_block,
                )

    def poll_blocks(
        self,
        stop_block: Optional[int] = None,
        required_confirmations: Optional[int] = None,
        new_block_timeout: Optional[int] = None,
    ) -> Iterator[BlockAPI]:
        # Wait half the time as the block time
        # to get data faster.
        block_time = self.network.block_time
        wait_time = block_time / 2

        # The timeout occurs when there is no chain activity
        # after a certain time.
        timeout = (
            (10.0 if self.network.is_dev else 50 * block_time)
            if new_block_timeout is None
            else new_block_timeout
        )

        # Only yield confirmed blocks.
        if required_confirmations is None:
            required_confirmations = self.network.required_confirmations

        @dataclass
        class YieldAction:
            hash: bytes
            number: int
            time: float

        # Pretend we _did_ yield the last confirmed item, for logic's sake.
        fake_last_block = self.get_block(self.web3.eth.block_number - required_confirmations)
        last_num = fake_last_block.number or 0
        last_hash = fake_last_block.hash or HexBytes(0)
        last: YieldAction = YieldAction(number=last_num, hash=last_hash, time=time.time())

        # A helper method for various points of ensuring we didn't receive a timeout.
        def assert_chain_activity():
            time_waiting = time.time() - last.time
            if time_waiting > timeout:
                raise ProviderError("Timed out waiting for next block.")

        # Begin the daemon.
        while True:
            # The next block we want is simply 1 after the last.
            next_block = last.number + 1
            head = self._get_latest_block()
            try:
                if head.number is None or head.hash is None:
                    raise ProviderError("Head block has no number or hash.")

                # Use an "adjusted" head, based on the required confirmations.
                adjusted_head = self.get_block(head.number - required_confirmations)
                if adjusted_head.number is None or adjusted_head.hash is None:
                    raise ProviderError("Adjusted head block has no number or hash.")

            except Exception:
                # TODO: I did encounter this sometimes in a re-org, needs better handling
                # and maybe bubbling up the block number/hash exceptions above.
                assert_chain_activity()
                continue

            if adjusted_head.number == last.number and adjusted_head.hash == last.hash:
                # The chain has not moved! Verify we have activity.
                assert_chain_activity()
                time.sleep(wait_time)
                continue

            elif adjusted_head.number < last.number or (
                adjusted_head.number == last.number and adjusted_head.hash != last.hash
            ):
                # Re-org detected! Error and catch up the chain.
                logger.error(
                    "Chain has reorganized since returning the last block. "
                    "Try adjusting the required network confirmations."
                )
                # Catch up the chain by setting the "next" to this tiny head.
                next_block = adjusted_head.number

                last.time = time.time()
                # NOTE: Drop down to code outside switch-of-ifs

            elif adjusted_head.number < next_block:
                # Wait for the next block.
                # But first, let's make sure the chain is still active.
                assert_chain_activity()
                time.sleep(wait_time)
                continue

            # NOTE: Should only get here if yielding blocks!
            #  Either because it is finally time or because a re-org allows us.
            for block_idx in range(next_block, adjusted_head.number + 1):
                block = self.get_block(block_idx)
                if block.number is None or block.hash is None:
                    raise ProviderError("Block has no number or hash.")

                # Set the last action, used for checking timeouts and re-orgs.
                last = YieldAction(number=block.number, hash=block.hash, time=time.time())

                yield block

                # This is the point at which the daemon will end,
                # provided the user passes in a `stop_block` arg.
                if stop_block is not None and block.number >= stop_block:
                    return

    def poll_logs(
        self,
        stop_block: Optional[int] = None,
        address: Optional["AddressType"] = None,
        topics: Optional[list[Union[str, list[str]]]] = None,
        required_confirmations: Optional[int] = None,
        new_block_timeout: Optional[int] = None,
        events: Optional[list["EventABI"]] = None,
    ) -> Iterator[ContractLog]:
        events = events or []
        if required_confirmations is None:
            required_confirmations = self.network.required_confirmations

        if stop_block is not None:
            if stop_block <= (self._get_latest_block().number or 0):
                raise ValueError("'stop' argument must be in the future.")

        for block in self.poll_blocks(stop_block, required_confirmations, new_block_timeout):
            if block.number is None:
                raise ValueError("Block number cannot be None")

            log_params: dict[str, Any] = {
                "start_block": block.number,
                "stop_block": block.number,
                "events": events,
            }
            if address is not None:
                log_params["addresses"] = [address]
            if topics is not None:
                log_params["topic_filter"] = topics

            log_filter = LogFilter(**log_params)
            yield from self.get_contract_logs(log_filter)

    def block_ranges(self, start: int = 0, stop: Optional[int] = None, page: Optional[int] = None):
        if stop is None:
            stop = self.chain_manager.blocks.height
        if page is None:
            page = self.block_page_size

        for start_block in range(start, stop + 1, page):
            stop_block = min(stop, start_block + page - 1)
            yield start_block, stop_block

    def get_contract_logs(self, log_filter: LogFilter) -> Iterator[ContractLog]:
        height = self.chain_manager.blocks.height
        start_block = log_filter.start_block
        stop_block_arg = log_filter.stop_block if log_filter.stop_block is not None else height
        stop_block = min(stop_block_arg, height)
        block_ranges = self.block_ranges(start_block, stop_block, self.block_page_size)

        def fetch_log_page(block_range):
            start, stop = block_range
            update = {"start_block": start, "stop_block": stop}
            page_filter = log_filter.model_copy(update=update)

            # NOTE: Using JSON mode since used as request data.
            filter_params = page_filter.model_dump(mode="json")
            logs = self.make_request("eth_getLogs", [filter_params])
            return self.network.ecosystem.decode_logs(logs, *log_filter.events)

        with ThreadPoolExecutor(self.concurrency) as pool:
            for page in pool.map(fetch_log_page, block_ranges):
                yield from page

    def prepare_transaction(self, txn: TransactionAPI) -> TransactionAPI:
        # NOTE: Use "expected value" for Chain ID, so if it doesn't match actual, we raise
        txn.chain_id = self.network.chain_id

        from ape_ethereum.transactions import StaticFeeTransaction, TransactionType

        txn_type = TransactionType(txn.type)
        if (
            txn_type in (TransactionType.STATIC, TransactionType.ACCESS_LIST)
            and isinstance(txn, StaticFeeTransaction)
            and txn.gas_price is None
        ):
            txn.gas_price = self.gas_price
        elif txn_type in (
            TransactionType.DYNAMIC,
            TransactionType.SHARED_BLOB,
            TransactionType.SET_CODE,
        ):
            if txn.max_priority_fee is None:
                txn.max_priority_fee = self.priority_fee

            if txn.max_fee is None:
                multiplier = self.network.base_fee_multiplier
                txn.max_fee = int(self.base_fee * multiplier + txn.max_priority_fee)

            # else: Assume user specified the correct amount or txn will fail and waste gas

        if txn_type is TransactionType.ACCESS_LIST and isinstance(txn, AccessListTransaction):
            if not txn.access_list:
                try:
                    txn.access_list = self.create_access_list(txn)
                except APINotImplementedError:
                    pass

        gas_limit = self.network.gas_limit if txn.gas_limit is None else txn.gas_limit
        if gas_limit in (None, "auto") or isinstance(gas_limit, AutoGasLimit):
            multiplier = (
                gas_limit.multiplier
                if isinstance(gas_limit, AutoGasLimit)
                else self.network.auto_gas_multiplier
            )
            if multiplier != 1.0:
                gas = min(int(self.estimate_gas_cost(txn) * multiplier), self.max_gas)
            else:
                gas = self.estimate_gas_cost(txn)

            txn.gas_limit = gas

        elif gas_limit == "max":
            txn.gas_limit = self.max_gas

        elif gas_limit is not None and isinstance(gas_limit, int):
            txn.gas_limit = gas_limit

        if txn.required_confirmations is None:
            txn.required_confirmations = self.network.required_confirmations
        elif not isinstance(txn.required_confirmations, int) or txn.required_confirmations < 0:
            raise TransactionError("'required_confirmations' must be a positive integer.")

        return txn

    def send_transaction(self, txn: TransactionAPI) -> ReceiptAPI:
        vm_err = None
        txn_data = None
        try:
            txn_hash = self._send_transaction(txn)
        except (Web3RPCError, Web3ContractLogicError) as err:
            vm_err = self.get_virtual_machine_error(
                err, txn=txn, set_ape_traceback=txn.raise_on_revert
            )
            if txn.raise_on_revert:
                raise vm_err from err
            else:
                txn_hash = to_hex(txn.txn_hash)

        required_confirmations = (
            txn.required_confirmations
            if txn.required_confirmations is not None
            else self.network.required_confirmations
        )
        txn_data = txn_data or txn.model_dump(by_alias=True, mode="json")

        # Signature is excluded from the model fields, so we have to include it manually.
        txn_data["signature"] = txn.signature

        manual_mining = not getattr(self, "auto_mine", True)
        if vm_err or manual_mining:
            receipt = self._create_receipt(
                block_number=-1,  # Not in a block.
                error=vm_err,
                required_confirmations=required_confirmations,
                status=TransactionStatusEnum.FAILING if vm_err else TransactionStatusEnum.NO_ERROR,
                txn_hash=txn_hash,
                **txn_data,
            )
        else:
            receipt = self.get_receipt(
                txn_hash,
                required_confirmations=required_confirmations,
                transaction=txn_data,
            )

        # NOTE: Ensure to cache even the failed receipts.
        # NOTE: Caching must happen before error enrichment.
        self.chain_manager.history.append(receipt)

        if receipt.failed:
            # For some reason, some nodes have issues with integer-types.
            if isinstance(txn_data.get("type"), int):
                txn_data["type"] = to_hex(txn_data["type"])

            # NOTE: For some reason, some providers have issues with
            #   `nonce`, it's not needed anyway.
            txn_data.pop("nonce", None)

            # Signature causes issues when making call (instead of tx)
            txn_data.pop("signature", None)

            # NOTE: Using JSON mode since used as request data.
            txn_params = cast(TxParams, txn_data)

            # Replay txn to get revert reason
            try:
                self.web3.eth.call(txn_params)
            except Exception as err:
                vm_err = self.get_virtual_machine_error(
                    err, txn=txn, set_ape_traceback=txn.raise_on_revert
                )
                receipt.error = vm_err
                if txn.raise_on_revert:
                    raise vm_err from err

            if txn.raise_on_revert:
                # If we get here, for some reason the tx-replay did not produce
                # a VM error.
                receipt.raise_for_status()

        if receipt.error:
            logger.error(receipt.error)

        return receipt

    def _send_transaction(self, txn: TransactionAPI) -> str:
        txn_hash = None
        if txn.sender is not None and txn.signature is None:
            # Missing signature, user likely trying to use an unlocked account.
            attempt_send = True
            if (
                self.network.is_dev
                and txn.sender not in self.account_manager.test_accounts._impersonated_accounts
            ):
                try:
                    self.account_manager.test_accounts.impersonate_account(txn.sender)
                except NotImplementedError:
                    # Unable to impersonate. Try sending as raw-tx.
                    attempt_send = False

            if attempt_send:
                # For some reason, some nodes have issues with integer-types.
                txn_data = {
                    k: to_hex(v) if isinstance(v, int) else v
                    for k, v in txn.model_dump(by_alias=True, mode="json").items()
                }
                tx_params = cast(TxParams, txn_data)
                txn_hash = to_hex(self.web3.eth.send_transaction(tx_params))
            # else: attempt raw tx

        if txn_hash is None:
            txn_hash = to_hex(self.web3.eth.send_raw_transaction(txn.serialize_transaction()))

        return txn_hash

    def _post_connect(self):
        # Register the console contract for trace enrichment
        self.chain_manager.contracts.cache_contract_type(
            CONSOLE_ADDRESS,
            console_contract,
            ecosystem_key=self.network.ecosystem.name,
            network_key=self.network.name,
        )

    def make_request(self, rpc: str, parameters: Optional[Iterable] = None) -> Any:
        return request_with_retry(lambda: self._make_request(rpc, parameters=parameters))

    def _make_request(self, rpc: str, parameters: Optional[Iterable] = None) -> Any:
        parameters = parameters or []
        try:
            result = self.web3.provider.make_request(RPCEndpoint(rpc), parameters)
        except HTTPError as err:
            if "method not allowed" in str(err).lower():
                raise APINotImplementedError(
                    f"RPC method '{rpc}' is not implemented by this node instance."
                )

            elif err.response.status_code == 429:
                raise  # Raise as-is so rate-limit handling picks it up.

            raise ProviderError(str(err)) from err

        if "error" in result:
            error = result["error"]
            message = (
                error["message"] if isinstance(error, dict) and "message" in error else str(error)
            )

            if (
                "does not exist/is not available" in str(message)
                or re.match(r"[m|M]ethod .*?not found", message)
                or message.startswith("Unknown RPC Endpoint")
                or "RPC Endpoint has not been implemented" in message
                or message == "unimplemented"
            ):
                raise APINotImplementedError(
                    f"RPC method '{rpc}' is not implemented by this node instance."
                )

            raise ProviderError(message)

        elif "result" in result:
            return result.get("result", {})

        return result

    def stream_request(self, method: str, params: Iterable, iter_path: str = "result.item"):
        if not (uri := self.http_uri):
            raise ProviderError("This provider has no HTTP URI and is unable to stream requests.")

        payload = {"jsonrpc": "2.0", "id": 1, "method": method, "params": params}
        results = ijson.sendable_list()
        coroutine = ijson.items_coro(results, iter_path)
        resp = requests.post(uri, json=payload, stream=True)
        resp.raise_for_status()

        for chunk in resp.iter_content(chunk_size=2**17):
            coroutine.send(chunk)
            yield from results
            del results[:]

    def create_access_list(
        self, transaction: TransactionAPI, block_id: Optional["BlockID"] = None
    ) -> list[AccessList]:
        """
        Get the access list for a transaction use ``eth_createAccessList``.

        Args:
            transaction (:class:`~ape.api.transactions.TransactionAPI`): The
              transaction to check.
            block_id (:class:`~ape.types.BlockID`): Optionally specify a block
              ID. Defaults to using the latest block.

        Returns:
            list[:class:`~ape_ethereum.transactions.AccessList`]
        """
        # NOTE: Using JSON mode since used in request data.
        tx_dict = transaction.model_dump(by_alias=True, mode="json", exclude={"chain_id"})
        tx_dict_converted = {}
        for key, val in tx_dict.items():
            if isinstance(val, int):
                # This RPC requires hex-str values.
                if val > 0:
                    tx_dict_converted[key] = to_hex(val)
                # else: 0-values cause problems.

            else:
                tx_dict_converted[key] = val

        if not tx_dict_converted.get("to") and tx_dict.get("data") in (None, "0x"):
            # Contract creation with no data, can skip.
            return []

        arguments: list = [tx_dict_converted]
        if block_id is not None:
            arguments.append(block_id)

        result = self.make_request("eth_createAccessList", arguments)
        return [AccessList.model_validate(x) for x in result.get("accessList", [])]

    def get_virtual_machine_error(self, exception: Exception, **kwargs) -> VirtualMachineError:
        txn = kwargs.get("txn")
        if isinstance(exception, Web3ContractLogicError):
            # This happens from `assert` or `require` statements.
            return self._handle_execution_reverted(exception, **kwargs)

        if not len(exception.args):
            return VirtualMachineError(base_err=exception, **kwargs)

        err_data = exception.args[0] if (hasattr(exception, "args") and exception.args) else None
        if isinstance(err_data, str) and "execution reverted" in err_data:
            return self._handle_execution_reverted(exception, **kwargs)

        elif not isinstance(err_data, dict):
            # Maybe it is a JSON-str.
            # NOTE: For some reason, it comes back with single quotes though.
            try:
                err_data = json.loads(str(err_data or "").replace("'", '"'))
            except Exception:
                return VirtualMachineError(base_err=exception, **kwargs)

        if not (err_msg := err_data.get("message")):
            return VirtualMachineError(base_err=exception, **kwargs)

        elif txn is not None and "nonce too low" in str(err_msg):
            txn = cast(TransactionAPI, txn)
            new_err_msg = f"Nonce '{txn.nonce}' is too low"
            return VirtualMachineError(
                new_err_msg, base_err=exception, code=err_data.get("code"), **kwargs
            )

        elif "out of gas" in str(err_msg) or "intrinsic gas too low" in str(err_msg):
            return OutOfGasError(code=err_data.get("code"), base_err=exception, **kwargs)

        return VirtualMachineError(str(err_msg), code=(err_data or {}).get("code"), **kwargs)

    def _handle_execution_reverted(
        self,
        exception: Union[Exception, str],
        txn: Optional[TransactionAPI] = None,
        trace: _TRACE_ARG = None,
        contract_address: Optional["AddressType"] = None,
        source_traceback: _SOURCE_TRACEBACK_ARG = None,
        set_ape_traceback: Optional[bool] = None,
    ) -> ContractLogicError:
        if hasattr(exception, "args") and len(exception.args) == 2:
            message = exception.args[0].replace("execution reverted: ", "")
            data = exception.args[1]
        else:
            message = str(exception).split(":")[-1].strip()
            data = None

        params: dict = {
            "trace": trace,
            "contract_address": contract_address,
            "source_traceback": source_traceback,
        }
        if set_ape_traceback is not None:
            params["set_ape_traceback"] = set_ape_traceback

        no_reason = message == "execution reverted"

        if isinstance(exception, Web3ContractLogicError) and no_reason:
            # Check for custom exception data and use that as the message instead.
            # This allows compiler exception enrichment to function.
            if data != "no data" and is_hex(data):
                message = add_0x_prefix(data)

            else:
                if trace is None and txn is not None:
                    trace = self.provider.get_transaction_trace(to_hex(txn.txn_hash))

                if trace is not None:
                    if callable(trace):
                        trace_called = params["trace"] = trace()
                    else:
                        trace_called = trace

                    if trace_called is not None and (revert_message := trace_called.revert_message):
                        message = revert_message
                        no_reason = False

        result = (
            ContractLogicError(txn=txn, **params)
            if no_reason
            else ContractLogicError(
                base_err=exception if isinstance(exception, Exception) else None,
                revert_message=message,
                txn=txn,
                **params,
            )
        )
        enriched = self.compiler_manager.enrich_error(result)

        # Show call trace if available
        if enriched.txn:
            # Unlikely scenario where a transaction is on the error even though a receipt exists.
            if isinstance(enriched.txn, TransactionAPI) and enriched.txn.receipt:
                enriched.txn.receipt.show_trace()
            elif isinstance(enriched.txn, ReceiptAPI):
                enriched.txn.show_trace()

        return enriched


class EthereumNodeProvider(Web3Provider, ABC):
    # optimal values for geth
    block_page_size: int = 5000
    concurrency: int = 16

    name: str = "node"

    # NOTE: Appends user-agent to base User-Agent string.
    request_header: dict = {"User-Agent": f"EthereumNodeProvider/web3.py/{web3_version}"}

    @property
    def connection_str(self) -> str:
        return self.uri or f"{self.ipc_path}"

    @property
    def connection_id(self) -> Optional[str]:
        return f"{self.network_choice}:{self.uri}"

    @property
    def _clean_uri(self) -> str:
        uri = f"{self.uri}"
        return sanitize_url(uri) if _is_http_url(uri) or _is_ws_url(uri) else uri

    @property
    def data_dir(self) -> Path:
        if self.settings.data_dir:
            return self.settings.data_dir.expanduser()

        return _get_default_data_dir()

    @property
    def ipc_path(self) -> Optional[Path]:
        if path := super().ipc_path:
            return path

        # Default (used by geth-process).
        return self.data_dir / "geth.ipc"

    @cached_property
    def has_poa_history(self) -> bool:
        """
        ``True`` if detected any PoA history. If the chain was _ever_ PoA, the special
        middleware is needed for web3.py. Provider plugins use this property when
        creating Web3 instances.
        """
        findings = False
        for option in ("earliest", "latest"):
            try:
                block = self.web3.eth.get_block(option)  # type: ignore[arg-type]

            except ExtraDataLengthError:
                findings = True
                break

            except Exception:
                # Some chains are "light" and we may not be able to detect
                # if it needs PoA middleware.
                continue

            else:
                findings = (
                    "proofOfAuthorityData" in block
                    or len(block.get("extraData", "")) > MAX_EXTRADATA_LENGTH
                )
                if findings:
                    break

        return findings

    @cached_property
    def _ots_api_level(self) -> Optional[int]:
        # NOTE: Returns None when OTS namespace is not enabled.
        try:
            result = self.make_request("ots_getApiLevel")
        except (NotImplementedError, ApeException, ValueError):
            return None

        if isinstance(result, int):
            return result

        elif isinstance(result, str) and result.isnumeric():
            return int(result)

        return None

    def _set_web3(self):
        # Clear cached version when connecting to another URI.
        self._client_version = None
        headers = self.network_manager.get_request_headers(
            self.network.ecosystem.name, self.network.name, self.name
        )
        self._web3 = _create_web3(
            http_uri=self.http_uri,
            ipc_path=self.ipc_path,
            ws_uri=self.ws_uri,
            request_kwargs={"headers": headers},
        )

    def _complete_connect(self):
        client_version = self.client_version.lower()
        if "geth" in client_version:
            self._log_connection("Geth")
        elif "reth" in client_version:
            self._log_connection("Reth")
        elif "erigon" in client_version:
            self._log_connection("Erigon")
            self.concurrency = 8
            self.block_page_size = 40_000
        elif "nethermind" in client_version:
            self._log_connection("Nethermind")
            self.concurrency = 32
            self.block_page_size = 50_000
        else:
            client_name = client_version.partition("/")[0]
            logger.info(f"Connecting to a '{client_name}' node.")

        if not self.network.is_dev:
            self.web3.eth.set_gas_price_strategy(rpc_gas_price_strategy)

        if self.has_poa_history and ExtraDataToPOAMiddleware not in self.web3.middleware_onion:
            self.web3.middleware_onion.inject(ExtraDataToPOAMiddleware, layer=0)

        chain_id = self.chain_id
        self.network.verify_chain_id(chain_id)

        # Correct network name, if using custom-URL approach.
        if self.network.name == "custom":
            for ecosystem_name, network in PUBLIC_CHAIN_META.items():
                for network_name, meta in network.items():
                    if "chainId" not in meta or meta["chainId"] != chain_id:
                        continue

                    # Network found.
                    self.network.name = network_name
                    self.network.ecosystem.name = ecosystem_name
                    break

    def disconnect(self):
        self._call_trace_approach = None
        self._web3 = None
        self._client_version = None

    def _log_connection(self, client_name: str):
        msg = f"Connecting to existing {client_name.strip()} node at"

        suffix = (
            self.ipc_path.as_posix().replace(Path.home().as_posix(), "$HOME")
            if self.ipc_path is not None and self.ipc_path.exists()
            else self._clean_uri
        )
        logger.info(f"{msg} {suffix}.")

    def ots_get_contract_creator(self, address: "AddressType") -> Optional[dict]:
        if self._ots_api_level is None:
            return None

        result = self.make_request("ots_getContractCreator", [address])
        if result is None:
            # NOTE: Skip the explorer part of the error message via `has_explorer=True`.
            raise ContractNotFoundError(address, True, self.network_choice)

        return result

    def _get_contract_creation_receipt(self, address: "AddressType") -> Optional[ReceiptAPI]:
        if result := self.ots_get_contract_creator(address):
            tx_hash = result["hash"]
            return self.get_receipt(tx_hash)

        return None

    def connect(self):
        self._set_web3()
        if not self.is_connected:
            uri = self._clean_uri
            message = f"No (supported) node found on '{uri}'."
            raise ProviderError(message)

        self._complete_connect()

    def simulate_transaction_bundle(
        self, bundle: Bundle, sim_overrides: Optional[dict] = None
    ) -> SimulationReport:
        """
        Submit a bundle and get the simulation result.

        Args:
            bundle (:class:`~ape.types.private_mempool.Bundle`) A bundle of transactions to send to the matchmaker.
            sim_overrides (dict | None) Optional fields to override simulation state.

        Returns:
            :class:`~ape.types.private_mempool.SimulationReport`
        """
        bundle_request = {"bundle": bundle.model_dump(), "simOverrides": sim_overrides or {}}
        result = self.provider.make_request("mev_simBundle", bundle_request)
        return SimulationReport.model_validate(result)


def _create_web3(
    http_uri: Optional[str] = None,
    ipc_path: Optional[Path] = None,
    ws_uri: Optional[str] = None,
    request_kwargs: Optional[dict] = None,
):
    # NOTE: This list is ordered by try-attempt.
    # Try ENV, then IPC, and then HTTP last.
    providers: list = [load_provider_from_environment]
    if ipc := ipc_path:
        providers.append(lambda: IPCProvider(ipc_path=ipc))
    if http := http_uri:
        request_kwargs = request_kwargs or {}
        if "timeout" not in request_kwargs:
            request_kwargs["timeout"] = 30 * 60

        providers.append(lambda: HTTPProvider(endpoint_uri=http, request_kwargs=request_kwargs))
    if ws := ws_uri:
        providers.append(lambda: WebsocketProvider(endpoint_uri=ws))

    provider = AutoProvider(potential_providers=providers)

    # TODO: Getting attribute error without this. Figure out why and do proper fix.
    class MockBatchingContext:
        def get(self, *args, **kwargs):
            return None

    provider._batching_context = MockBatchingContext()  # type: ignore

    return Web3(provider, middleware=[])


def _get_default_data_dir() -> Path:
    # Modified from web3.py package to always return IPC even when none exist.
    if sys.platform == "darwin":
        return Path.home() / "Library" / "Ethereum"

    elif sys.platform.startswith("linux") or sys.platform.startswith("freebsd"):
        return Path.home() / "ethereum"

    elif sys.platform == "win32":
        return Path(os.path.join("\\\\", ".", "pipe"))

    else:
        raise ValueError(
            f"Unsupported platform '{sys.platform}'.  Only darwin/linux/win32/"
            "freebsd are supported.  You must specify the data_dir."
        )


def _is_uri(val: str) -> bool:
    return _is_http_url(val) or _is_ws_url(val) or _is_ipc_path(val)


def _is_http_url(val: str) -> bool:
    return val.startswith("https://") or val.startswith("http://")


def _is_ws_url(val: str) -> bool:
    return val.startswith("wss://") or val.startswith("ws://")


def _is_ipc_path(val: Union[str, Path]) -> bool:
    return f"{val}".endswith(".ipc")


class _LazyCallTrace(ManagerAccessMixin):
    def __init__(self, eth_call_args: list):
        self._arguments = eth_call_args

        self.contract_type = None

    @cached_property
    def trace(self) -> CallTrace:
        return CallTrace(
            tx=self._arguments[0], arguments=self._arguments[1:], use_tokens_for_symbols=True
        )

    @cached_property
    def source_traceback(self) -> Optional[SourceTraceback]:
        ct = self.contract_type
        if ct is None:
            return None

        method_id = self._arguments[0].get("data", "")[:10] or None
        if ct and method_id:
            if contract_src := self.local_project._create_contract_source(ct):
                return SourceTraceback.create(contract_src, self.trace, method_id)

        return None



import re
import warnings
from typing import Optional

import pytest
from eth_pydantic_types import HexBytes
from eth_utils import to_hex
from hexbytes import HexBytes as BaseHexBytes

from ape.api import TransactionAPI
from ape.exceptions import SignatureError
from ape_ethereum.transactions import (
    AccessList,
    AccessListTransaction,
    DynamicFeeTransaction,
    StaticFeeTransaction,
    TransactionType,
)

ACCESS_LIST = [{"address": "0x0000000000000000000000000000000000000004", "storageKeys": []}]

ACCESS_LIST_HEXBYTES = [
    {
        "address": "0x0000000000000000000000000000000000000004",
        "storageKeys": [
            BaseHexBytes("0x0000000000000000000000000000000000000000000000000000000000000000")
        ],
    }
]

# NOTE: Long access List also uses / tests HexBytes from eth_pydantic_types\
#   (which shouldn't matter).
LONG_ACCESS_LIST = [
    {
        "address": "0x5Ab952D45d33ba32DBAA3Da85b0738aC9DF24626",
        "storageKeys": [
            HexBytes("0x0000000000000000000000000000000000000000000000000000000000000000"),
            HexBytes("0x0000000000000000000000000000000000000000000000000000000000000004"),
            HexBytes("0x0000000000000000000000000000000000000000000000000000000000000001"),
            HexBytes("0x9c04773acff4c5c42718bd0120c72761f458e43068a3961eb935577d1ed4effb"),
            HexBytes("0x0000000000000000000000000000000000000000000000000000000000000008"),
        ],
    },
    {
        "address": "0x60594a405d53811d3BC4766596EFD80fd545A270",
        "storageKeys": [
            HexBytes("0x2e7eb34672b18adb9244b4932b747ae4bcb4f839a51cd3b2e72e6113c9d4d285"),
            HexBytes("0x000000000000000000000000000000000000000000000000000000000000005c"),
            HexBytes("0x000000000000000000000000000000000000000000000000000000000000005d"),
            HexBytes("0x0000000000000000000000000000000000000000000000000000000000000000"),
            HexBytes("0x0000000000000000000000000000000000000000000000000000000000000004"),
            HexBytes("0x0000000000000000000000000000000000000000000000000000000000000002"),
        ],
    },
    {
        "address": "0x6B175474E89094C44Da98b954EedeAC495271d0F",
        "storageKeys": [
            HexBytes("0x29cb8bd4e192d16f51155329ce8b0f5eb88a1d9e4d3b93ce07efbac9e1c4d175"),
            HexBytes("0x995f3b129dd3291868ddb9cf202c75cd985227d50e309847fbab0f8da403b19c"),
        ],
    },
    {
        "address": "0x709b80deC74CA88e9394DEaB45840d861BD5398d",
        "storageKeys": [
            HexBytes("0x0000000000000000000000000000000000000000000000000000000000000006"),
            HexBytes("0x0000000000000000000000000000000000000000000000000000000000000007"),
            HexBytes("0x0000000000000000000000000000000000000000000000000000000000000009"),
            HexBytes("0x000000000000000000000000000000000000000000000000000000000000000a"),
            HexBytes("0x000000000000000000000000000000000000000000000000000000000000000c"),
            HexBytes("0x0000000000000000000000000000000000000000000000000000000000000008"),
        ],
    },
    {
        "address": "0x45E412E1878080815D6D51d47B83D17869433459",
        "storageKeys": [
            HexBytes("0x0000000000000000000000000000000000000000000000000000000000000005"),
            HexBytes("0x0000000000000000000000000000000000000000000000000000000000000008"),
            HexBytes("0x3a8870f40d200d252a256529d1f22c4f977b6e09dc78c6ebdeffd3f8651c4719"),
            HexBytes("0x2d027208e382c51007968756a0988c06fc25a78a1ef13c193023db7354fa6ea9"),
            HexBytes("0xc247e5713292da7b6b8145ca699e5c90c1257a929a9b107aa7c7d211bc3a369c"),
            HexBytes("0x50164d12bd200c7817ac53416fd234974ec726930dca7760174d351e0e29af6a"),
            HexBytes("0x000000000000000000000000000000000000000000000000000000000000000f"),
            HexBytes("0xb8e6a40a2729c4eca72da15a47bea21de47ee5b868eb98a2a3966fe05bc777c6"),
            HexBytes("0x1e4e2ca44a4ccf068a5ab14ea9a4d61e97a8b5395bf782d9b2032e0ee8487c29"),
            HexBytes("0x000000000000000000000000000000000000000000000000000000000000000b"),
            HexBytes("0x5de84563fdc81a8663ce72466e3e1e667da5e6c8834c90c011812476ee214f3c"),
            HexBytes("0xb39e9ba92c3c47c76d4f70e3bc9c3270ab78d2592718d377c8f5433a34d3470a"),
            HexBytes("0x86853267f2534b9eec92fd7a80680d7481e5e740ce8d21c54361341b3a3c1f14"),
            HexBytes("0xa78b521343fce79b129d9bbe9bff921a08c8a8fde6ae24a7e159847b3ba54bb7"),
            HexBytes("0x0000000000000000000000000000000000000000000000000000000000000010"),
            HexBytes("0x000000000000000000000000000000000000000000000000000000000000000a"),
            HexBytes("0x000000000000000000000000000000000000000000000000000000000000001d"),
            HexBytes("0x0000000000000000000000000000000000000000000000000000000000000009"),
            HexBytes("0x21c632ec26149a6f0282e6b24bf8cb49cdf334cfb6e21a8f4001c5e39c858863"),
        ],
    },
    {
        "address": "0xFb76CD5e9Fb9351137C3CE0aC1C23212C46995A7",
        "storageKeys": [
            HexBytes("0x0000000000000000000000000000000000000000000000000000000000000002"),
            HexBytes("0x9c04773acff4c5c42718bd0120c72761f458e43068a3961eb935577d1ed4effb"),
            HexBytes("0x0000000000000000000000000000000000000000000000000000000000000008"),
            HexBytes("0x0000000000000000000000000000000000000000000000000000000000000000"),
            HexBytes("0x0000000000000000000000000000000000000000000000000000000000000004"),
        ],
    },
    {
        "address": "0x3005003BDA885deE7c74182e5FE336e9E3Df87bB",
        "storageKeys": [
            HexBytes("0x23d376ed44ad443baf938cf24280e50027474804f2cdc01ec9195cdc123467a5"),
            HexBytes("0x577b913a3c8810dd10161c9ae11e2ee31042564c62114c83b0bc5d3a3e71b362"),
            HexBytes("0x04421adc6f4dd2e52160b542e3ea06d21bd261fae596660765cf3964443fa2b7"),
        ],
    },
    {
        "address": "0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2",
        "storageKeys": [
            HexBytes("0xbb0d37a1af2c606132ece9f63118a24cfb959e4791cb65d639217bd82f7282b3"),
            HexBytes("0xf762dfe765e313d39f5dd6e34e29a9ef0af51578e67f7f482bb4f8efd984976b"),
            HexBytes("0xa3303497cfb83efa67ca4f88b647c0dc6639b10da44cd96ac8b3488d155cf935"),
            HexBytes("0xb73247ebb5deb7b0fcdb78323c538aca3746ddf79808af84f3cf760fcb915185"),
            HexBytes("0x12231cd4c753cb5530a43a74c45106c24765e6f81dc8927d4f4be7e53315d5a8"),
        ],
    },
]


@pytest.fixture
def access_list():
    return ACCESS_LIST


@pytest.mark.parametrize("type_kwarg", (0, "0x0", b"\x00", "0", HexBytes("0x0"), HexBytes("0x00")))
def test_type_0_transactions(ethereum, type_kwarg):
    txn = ethereum.create_transaction(type=type_kwarg)
    assert txn.type == TransactionType.STATIC.value


@pytest.mark.parametrize("type_kwarg", (1, "0x01", b"\x01", "1", "01", HexBytes("0x01")))
def test_type_1_transactions(ethereum, type_kwarg):
    txn = ethereum.create_transaction(type=type_kwarg)
    assert txn.type == TransactionType.ACCESS_LIST.value


@pytest.mark.parametrize("type_kwarg", (None, 2, "0x02", b"\x02", "2", "02", HexBytes("0x02")))
def test_type_2_transactions(ethereum, type_kwarg):
    txn = ethereum.create_transaction(type=type_kwarg)
    assert txn.type == TransactionType.DYNAMIC.value


@pytest.mark.parametrize("key", ("accessList", "access_list"))
def test_type_1_transactions_using_access_list(ethereum, access_list, key):
    """
    If not given type and only given accessList, the assumed type is 1,
    an "access-list" transaction.
    """
    data = {key: access_list}
    txn = ethereum.create_transaction(**data)
    assert txn.type == 1


@pytest.mark.parametrize("key", ("accessList", "access_list"))
def test_type_2_transactions_with_max_fee_and_access_list(ethereum, access_list, key):
    """
    Dynamic-fee txns also support access lists, so the presence of max_fee
    with access_list implies a type 2 txn.
    """
    data = {"max_fee": 1000000000, key: access_list}
    txn = ethereum.create_transaction(**data)
    assert txn.type == 2
    assert txn.max_fee == 1000000000


def test_type_2_transactions_with_access_list(ethereum, access_list):
    txn = ethereum.create_transaction(type=2, accessList=access_list)
    assert txn.type == TransactionType.DYNAMIC.value
    assert txn.access_list == [AccessList.model_validate(x) for x in access_list]


@pytest.mark.parametrize(
    "tx_kwargs",
    [{"type": 3}, {"max_fee_per_blob_gas": 123}, {"blob_versioned_hashes": [HexBytes(123)]}],
)
def test_type_3_transactions(ethereum, tx_kwargs):
    txn = ethereum.create_transaction(**tx_kwargs)
    assert txn.type == 3


@pytest.mark.parametrize(
    "tx_kwargs",
    [{"type": 4}, {"authorizationList": []}],
)
def test_type_4_transactions(ethereum, tx_kwargs):
    txn = ethereum.create_transaction(**tx_kwargs)
    assert txn.type == 4


@pytest.mark.parametrize(
    "fee_kwargs",
    (
        {"max_fee": "100 gwei"},
        {"max_fee": int(100e9)},
        {"max_priority_fee": "1 gwei"},
        {"max_priority_fee": int(1e9)},
        {"max_priority_fee": "1 gwei", "max_fee": "100 gwei"},
        {"max_priority_fee": int(1e9), "max_fee": "100 gwei"},
        {"max_priority_fee": "1 gwei", "max_fee": int(100e9)},
        {"max_priority_fee": int(1e9), "max_fee": int(100e9)},
    ),
)
def test_type_2_transactions_using_fee_kwargs(ethereum, fee_kwargs):
    txn = ethereum.create_transaction(**fee_kwargs)
    assert isinstance(txn, DynamicFeeTransaction)
    if "max_priority_fee" in fee_kwargs:
        assert txn.max_priority_fee == int(1e9)
    if "max_fee" in fee_kwargs:
        assert txn.max_fee == int(100e9)


@pytest.mark.parametrize(
    "kwargs",
    [
        {"type": 0},
        {"type": 1},
        {"type": 1, "accessList": ACCESS_LIST},
        {"type": 1, "accessList": LONG_ACCESS_LIST},
        {"type": 2, "max_fee": 1_000_000_000},
        {"type": 2, "accessList": ACCESS_LIST},
        {"type": 2, "accessList": ACCESS_LIST_HEXBYTES},
    ],
)
def test_txn_hash_and_receipt(owner, eth_tester_provider, ethereum, kwargs):
    txn = ethereum.create_transaction(**kwargs)
    txn = owner.prepare_transaction(txn)
    txn = owner.sign_transaction(txn)
    assert txn

    # Show the .hash alias works.
    assert txn.hash == txn.txn_hash

    actual = to_hex(txn.txn_hash)
    receipt = eth_tester_provider.send_transaction(txn)

    # Show that we can access the receipt from the transaction.
    assert txn.receipt == receipt

    expected = receipt.txn_hash

    assert actual == expected


def test_txn_hash_when_access_list_is_raw(ethereum, owner):
    """
    Tests against a condition I was never able to reproduce where
    a transaction's access list contained bytes-values and that
    causes the serialization to error.
    """

    txn = ethereum.create_transaction(accessList=ACCESS_LIST_HEXBYTES, type=2)
    txn = owner.prepare_transaction(txn)
    txn = owner.sign_transaction(txn)

    # Hack to make access_list raw. I am not sure how a user would get
    # to this state, but somehow they have.
    txn.access_list = ACCESS_LIST_HEXBYTES

    # Ignore the Pydantic warning from access-list being the wrong type.
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        actual = to_hex(txn.txn_hash)

    assert actual.startswith("0x")


def test_data_when_contains_whitespace():
    data = b"Should not clip whitespace\t\n"
    txn_dict = {"data": data}
    txn = StaticFeeTransaction.model_validate(txn_dict)
    assert txn.data == data, "Whitespace should not be removed from data"


def test_model_dump_excludes_none_values():
    txn = StaticFeeTransaction(sender=None)
    txn.value = 1000000
    actual = txn.model_dump()
    assert "value" in actual
    txn.value = None  # type: ignore
    actual = txn.model_dump()
    assert "value" not in actual


def test_model_dump_access_list():
    # Data directly from eth_createAccessList RPC
    access_list = [
        {
            "address": "0x7ef8e99980da5bcedcf7c10f41e55f759f6a174b",
            "storageKeys": [
                "0x0000000000000000000000000000000000000000000000000000000000000000",
                "0x0000000000000000000000000000000000000000000000000000000000000001",
                "0x0000000000000000000000000000000000000000000000000000000000000002",
            ],
        }
    ]
    txn = AccessListTransaction(access_list=access_list, sender=None)
    actual = txn.model_dump(exclude_none=True, by_alias=True)
    assert actual is not None


def test_str_when_data_is_bytes(ethereum):
    """
    Tests against a condition that would cause transactions to
    fail with string-encoding errors.
    """
    txn = ethereum.create_transaction(data=HexBytes("0x123"))
    actual = str(txn)
    assert isinstance(actual, str)


def test_str_when_data_is_long_shows_first_4_bytes(vyper_contract_instance):
    """
    Tests against a condition that would cause transactions to
    fail with string-encoding errors.
    """
    txn = vyper_contract_instance.setNumber.as_transaction(123)
    actual = str(txn)
    assert isinstance(actual, str)
    assert "data: 0x3fb5c1cb..." in actual


def test_str_when_data_is_long_and_configured_full_calldata(project, vyper_contract_instance):
    txn = vyper_contract_instance.setNumber.as_transaction(123)
    with project.temp_config(display={"calldata": "full"}):
        actual = str(txn)

    expected = "data: 0x3fb5c1cb000000000000000000000000000000000000000000000000000000000000007b"
    assert isinstance(actual, str)
    assert expected in actual


def test_receipt_when_none(ethereum):
    txn = ethereum.create_transaction(data=HexBytes("0x123"))
    assert txn.receipt is None


def test_repr(ethereum):
    txn = ethereum.create_transaction(data=HexBytes("0x123"))
    actual = repr(txn)
    expected = (
        r"<DynamicFeeTransaction chainId=\d*, "
        r"gas=\d*, value=0, data=0x\d*, type=2, accessList=\[\]>"
    )
    assert re.match(expected, actual)


# NOTE: Some of these values are needed for signing to work.
@pytest.mark.parametrize(
    "tx_kwargs",
    [
        {"data": HexBytes("0x123"), "nonce": 0, "gas_price": 0},
        {"gasLimit": 100, "nonce": 0, "max_fee": 0, "max_priority_fee": 0},
        {"access_list": ACCESS_LIST, "nonce": 0, "gasPrice": 0},  # NOTE: show camelCase works
        {
            "accessList": ACCESS_LIST_HEXBYTES,
            "nonce": 0,
            "max_fee": 0,
            "max_priority_fee": 0,
            "type": 2,
        },
        {
            "access_list": LONG_ACCESS_LIST,
            "nonce": 0,
            "max_fee": 0,
            "max_priority_fee": 0,
            "type": 2,
        },
    ],
)
def test_serialize_transaction(owner, ethereum, tx_kwargs):
    txn = ethereum.create_transaction(**tx_kwargs)
    txn = owner.sign_transaction(txn)
    assert txn is not None

    actual = txn.serialize_transaction()
    assert isinstance(actual, bytes)


def test_serialize_transaction_missing_signature(ethereum, owner):
    expected = r"The transaction is not signed."
    txn = ethereum.create_transaction(data=HexBytes("0x123"), sender=owner.address)
    with pytest.raises(SignatureError, match=expected):
        txn.serialize_transaction()


def test_serialize_transaction_missing_signature_and_sender(ethereum):
    expected = (
        r"The transaction is not signed. "
        r"Did you forget to add the `sender=` kwarg to the transaction function call?"
    )
    txn = ethereum.create_transaction(data=HexBytes("0x123"))
    with pytest.raises(SignatureError, match=expected):
        txn.serialize_transaction()


class TestAccessList:
    @pytest.mark.parametrize(
        "address",
        (
            "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
            HexBytes("0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"),
        ),
    )
    def test_address(self, address):
        actual = AccessList(address=address, storageKeys=[])
        assert actual.address == "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"

    @pytest.mark.parametrize("storage_key", (123, HexBytes(123), "0x0123"))
    def test_storage_keys(self, storage_key, zero_address):
        actual = AccessList(address=zero_address, storageKeys=[storage_key])
        assert actual.storage_keys == [HexBytes(storage_key)]


def test_override_annotated_fields():
    """
    This test is to prove that a user may use an `int` for a base-class
    when the API field is described as a `HexInt`.
    """

    class MyTransaction(TransactionAPI):
        @property
        def txn_hash(self) -> HexBytes:
            return HexBytes("")

        def serialize_transaction(self) -> bytes:
            return b""

        chain_id: Optional[int] = None  # The base type is `Optional[HexInt]`.

    chain_id = 123123123123123123123123123123
    tx_type = 120
    my_tx = MyTransaction.model_validate({"chain_id": chain_id, "type": tx_type})
    assert my_tx.chain_id == chain_id
    assert my_tx.type == tx_type


def test_gas(ethereum):
    tx = ethereum.create_transaction(gas_limit=123)
    assert tx.gas_limit == 123
    # Show the `gas` alias works.
    assert tx.gas == 123



from pathlib import Path

import pytest

from ape.utils.misc import SOURCE_EXCLUDE_PATTERNS
from ape.utils.os import (
    clean_path,
    create_tempdir,
    get_all_files_in_directory,
    get_full_extension,
    get_relative_path,
    path_match,
    run_in_tempdir,
)

_TEST_DIRECTORY_PATH = Path("/This/is/a/test/")
_TEST_FILE_PATH = _TEST_DIRECTORY_PATH / "scripts" / "script.py"


def test_get_relative_path_from_project():
    actual = get_relative_path(_TEST_FILE_PATH, _TEST_DIRECTORY_PATH)
    expected = Path("scripts/script.py")
    assert actual == expected


def test_get_relative_path_same_path():
    actual = get_relative_path(_TEST_FILE_PATH, _TEST_FILE_PATH)
    assert actual == Path()


def test_get_relative_path_roots():
    root = Path("/")
    actual = get_relative_path(root, root)
    assert actual == Path()


def test_get_all_files_in_directory():
    with create_tempdir() as path:
        first_dir = path / "First"
        second_dir = path / "Second"
        second_first_dir = second_dir / "SecondFirst"
        dirs = (first_dir, second_dir, second_first_dir)
        for dir_ in dirs:
            dir_.mkdir()

        file_a = first_dir / "test.t.txt"
        file_b = second_dir / "test.txt"
        file_c = second_dir / "test2.txt"
        file_d = second_first_dir / "test.inner.txt"
        file_e = second_first_dir / "test3.txt"
        files = (file_a, file_b, file_c, file_d, file_e)
        for file in files:
            file.touch()

        all_files = get_all_files_in_directory(path)
        txt_files = get_all_files_in_directory(path, pattern=r"\w+\.txt", max_files=2)
        t_txt_files = get_all_files_in_directory(path, pattern=r"\w+\.t.txt")
        inner_txt_files = get_all_files_in_directory(path, pattern=r"\w+\.inner.txt")

        assert len(all_files) == 5
        assert len(txt_files) == 2
        assert len(t_txt_files) == 1
        assert len(inner_txt_files) == 1


@pytest.mark.parametrize(
    "path,expected",
    [
        ("test.json", ".json"),
        ("Contract.t.sol", ".t.sol"),
        ("this.is.a.lot.of.dots", ".is.a.lot.of.dots"),
        ("directory", ""),
        (".privateDirectory", ""),
        ("path/to/.gitkeep", ""),
        (".config.json", ".json"),
    ],
)
def test_get_full_extension(path, expected):
    path = Path(path)
    actual = get_full_extension(path)
    assert actual == expected


@pytest.mark.parametrize("name", (None, "oogabooga"))
def test_create_tempdir(name):
    with create_tempdir(name=name) as path:
        assert isinstance(path, Path)
        assert path.resolve() == path

    assert not path.is_dir()
    if name is not None:
        assert path.name == name


@pytest.mark.parametrize("name", (None, "oogabooga"))
def test_run_in_tempdir(name):
    arguments = []  # using list to make global ref work.

    def fn(p):
        arguments.append(p)

    run_in_tempdir(fn, name=name)
    assert len(arguments) == 1  # did run
    assert isinstance(arguments[0], Path)
    assert arguments[0].resolve() == arguments[0]
    if name is not None:
        assert arguments[0].name == name


@pytest.mark.parametrize(
    "path",
    (
        "path/to/.cache/file/1.json",
        Path("path/to/.cache/file/1.json"),
        ".cache",
        Path(".cache"),
        Path("contracts/.cache/TEST.json"),
    ),
)
def test_path_match_cache_folder(path):
    assert path_match(path, *SOURCE_EXCLUDE_PATTERNS)


@pytest.mark.parametrize(
    "path",
    (
        "path/to/.build/__local__.json",
        Path("path/to/.build/__local__.json"),
        ".build",
        Path(".build"),
        Path("contracts/.build/TEST.json"),
    ),
)
def test_path_match_build_folder(path):
    assert path_match(path, *SOURCE_EXCLUDE_PATTERNS)


@pytest.mark.parametrize(
    "path",
    (
        "path/to/MyFile.build.json/__local__.json",
        Path("path/to/MyFile.build.json/__local__.json"),
        ".builder",
        Path(".cacher"),
        Path("contracts/.builder/TEST.json"),
    ),
)
def test_pat_match_does_not_match_but_close(path):
    """
    Ensures we don't get false positives.
    """
    assert not path_match(path, *SOURCE_EXCLUDE_PATTERNS)


@pytest.mark.parametrize(
    "path",
    (
        ".gitkeep",
        "path/to/.gitkeep",
        "index.html",
        "path/to/index.css",
        "py.typed",
    ),
)
def test_path_match(path):
    assert path_match(path, *SOURCE_EXCLUDE_PATTERNS)


@pytest.mark.parametrize(
    "path,exc",
    [
        ("path/to/MyContract.t.sol", "*t.sol"),
        ("contracts/mocks/MyContract.sol", "mocks"),
        ("mocks/MyContract.sol", "mocks"),
        ("MockContract.sol", "Mock*"),
    ],
)
def test_path_match_test_contracts(path, exc):
    assert not path_match(path, *SOURCE_EXCLUDE_PATTERNS)
    exclusions = [*SOURCE_EXCLUDE_PATTERNS, exc]
    assert path_match(path, *exclusions)


@pytest.mark.parametrize(
    "path",
    (
        Path("path/to/contracts/exclude_dir/subdir/MyContract.json"),
        Path("exclude_dir/subdir/MyContract.json"),
        Path("exclude_dir/MyContract.json"),
    ),
)
def test_path_match_recurse_dir(path):
    """
    Testing a specific way of excluding all the files in a directory.
    """
    excl = "exclude_dir/**"
    assert path_match(path, excl)


def test_clean_path_relative_to_home():
    name = "__canary_ape_test__"
    path = Path.home() / name
    actual = clean_path(path)
    expected = f"$HOME/{name}"
    assert actual == expected


def test_clean_path_not_relative_to_home():
    name = "__canary_ape_test__"
    path = Path(name)
    actual = clean_path(path)
    assert actual == name



import json
import os
import re
import stat
import sys
import tarfile
import zipfile
from collections.abc import Callable, Iterator
from contextlib import contextmanager
from fnmatch import fnmatch
from importlib.metadata import PackageNotFoundError, distribution
from pathlib import Path
from re import Pattern
from tempfile import TemporaryDirectory, gettempdir
from typing import Any, Optional, Union


# TODO: This method is no longer needed since the dropping of 3.9
#   Delete in Ape 0.9 release.
def is_relative_to(path: Path, target: Path) -> bool:
    """
    Search a path and determine its relevancy.

    Args:
        path (str): Path represents a filesystem to find.
        target (str): Path represents a filesystem to match.

    Returns:
        bool: ``True`` if the path is relative to the target path or ``False``.
    """
    return target.is_relative_to(path)


def get_relative_path(target: Path, anchor: Path) -> Path:
    """
    Compute the relative path of ``target`` relative to ``anchor``,
    which may or may not share a common ancestor.

    **NOTE ON PERFORMANCE**: Both paths must be absolute to
    use this method. If you know both methods are absolute,
    this method is a performance boost. If you have to first
    call `.absolute()` on the paths, use
    `target.relative_to(anchor)` instead; as it will be
    faster in that case.

    Args:
        target (pathlib.Path): The path we are interested in.
        anchor (pathlib.Path): The path we are starting from.

    Returns:
        pathlib.Path: The new path to the target path from the anchor path.
    """
    # Calculate common prefix length
    common_parts = 0
    for target_part, anchor_part in zip(target.parts, anchor.parts):
        if target_part == anchor_part:
            common_parts += 1
        else:
            break

    # Calculate the relative path
    relative_parts = [".."] * (len(anchor.parts) - common_parts) + list(target.parts[common_parts:])
    return Path(*relative_parts)


def get_all_files_in_directory(
    path: Path, pattern: Optional[Union[Pattern, str]] = None, max_files: Optional[int] = None
) -> list[Path]:
    """
    Returns all the files in a directory structure (recursive).

    For example, given a directory structure like::

        dir_a: dir_b, file_a, file_b
        dir_b: file_c

    and you provide the path to ``dir_a``, it will return a list containing
    the paths to ``file_a``, ``file_b`` and ``file_c``.

    Args:
        path (pathlib.Path): A directory containing files of interest.
        pattern (Optional[Union[Pattern, str]]): Optionally provide a regex
          pattern to match.
        max_files (Optional[int]): Optionally set a max file count. This is useful
          because huge file structures will be very slow.

    Returns:
        list[pathlib.Path]: A list of files in the given directory.
    """
    if path.is_file():
        return [path]
    elif not path.is_dir():
        return []

    pattern_obj: Optional[Pattern] = None
    if isinstance(pattern, str):
        pattern_obj = re.compile(pattern)
    elif pattern is not None:
        pattern_obj = pattern

    result: list[Path] = []
    append_result = result.append  # Local variable for faster access
    for file in path.rglob("*.*"):
        if not file.is_file() or (pattern_obj is not None and not pattern_obj.match(file.name)):
            continue

        append_result(file)
        if max_files is not None and len(result) >= max_files:
            break

    return result


def expand_environment_variables(contents: str) -> str:
    """
    Replace substrings of the form ``$name`` or ``${name}`` in the given path
    with the value of environment variable name.

    Args:
        contents (str): A path-like object representing a file system.
                            A path-like object is either a string or bytes object
                            representing a path.
    Returns:
        str: The given content with all environment variables replaced with their values.
    """
    return os.path.expandvars(contents)


class use_temp_sys_path:
    """
    A context manager to manage injecting and removing paths from
    a user's sys paths without permanently modifying it.
    """

    def __init__(self, path: Path, exclude: Optional[list[Path]] = None):
        self.temp_path = str(path)
        self.exclude = [str(p) for p in exclude or []]

    def __enter__(self):
        for path in self.exclude:
            if path in sys.path:
                sys.path.remove(path)
            else:
                # Preventing trying to re-add during exit.
                self.exclude.remove(path)

        if self.temp_path not in sys.path:
            sys.path.append(self.temp_path)

    def __exit__(self, *exc):
        if self.temp_path in sys.path:
            sys.path.remove(self.temp_path)

        for path in self.exclude:
            if path not in sys.path:
                sys.path.append(path)


def get_full_extension(path: Union[Path, str]) -> str:
    """
    For a path like ``Path("Contract.t.sol")``,
    returns ``.t.sol``, unlike the regular Path
    property ``.suffix`` which returns ``.sol``.

    Args:
        path (Path | str): The path with an extension.

    Returns:
        str: The full suffix
    """
    if not path:
        return ""

    path = Path(path)
    if path.is_dir() or path.suffix == "":
        return ""

    name = path.name
    parts = name.split(".")

    if len(parts) > 2 and name.startswith("."):
        return "." + ".".join(parts[2:])
    elif len(parts) > 1:
        return "." + ".".join(parts[1:])

    return ""


@contextmanager
def create_tempdir(name: Optional[str] = None) -> Iterator[Path]:
    """
    Create a temporary directory. Differs from ``TemporaryDirectory()``
    context-call alone because it automatically resolves the path.

    Args:
        name (Optional[str]): Optional provide a name of  the directory.
          Else, defaults to root of ``tempfile.TemporaryDirectory()``
          (resolved).

    Returns:
        Iterator[Path]: Context managing the temporary directory.
    """
    with TemporaryDirectory() as temp_dir:
        temp_path = Path(temp_dir).resolve()

        if name:
            path = temp_path / name
            path.mkdir()
        else:
            path = temp_path

        yield path


def run_in_tempdir(
    fn: Callable[[Path], Any],
    name: Optional[str] = None,
):
    """
    Run the given function in a temporary directory with its path
    resolved.

    Args:
        fn (Callable): A function that takes a path. It gets called
          with the resolved path to the temporary directory.
        name (Optional[str]): Optionally name the temporary directory.

    Returns:
        Any: The result of the function call.
    """
    with create_tempdir(name=name) as temp_dir:
        return fn(temp_dir)


def in_tempdir(path: Path) -> bool:
    """
    Returns ``True`` when the given path is in a temporary directory.

    Args:
        path (Path): The path to check.

    Returns:
        bool
    """
    temp_dir = os.path.normpath(f"{Path(gettempdir()).resolve()}")
    normalized_path = os.path.normpath(path)
    return normalized_path.startswith(temp_dir)


def path_match(path: Union[str, Path], *exclusions: str) -> bool:
    """
    A better glob-matching function. For example:

    >>> from pathlib import Path
    >>> p = Path("test/to/.build/me/2/file.json")
    >>> p.match("**/.build/**")
    False
    >>> from ape.utils.os import path_match
    >>> path_match(p, "**/.build/**")
    True
    """
    path_str = str(path)
    path_path = Path(path)

    for excl in exclusions:
        if fnmatch(path_str, excl):
            return True

        elif fnmatch(path_path.name, excl):
            return True

        else:
            # If the exclusion is he full name of any of the parents
            # (e.g. ".cache", it is a match).
            for parent in path_path.parents:
                if parent.name == excl:
                    return True

                # Walk the path recursively.
                relative_str = path_str.replace(str(parent), "").strip(os.path.sep)
                if fnmatch(relative_str, excl):
                    return True

    return False


def clean_path(path: Path) -> str:
    """
    Replace the home directory with key ``$HOME`` and return
    the path as a str. This is used for outputting paths
    with less doxxing.

    Args:
        path (Path): The path to sanitize.

    Returns:
        str: A sanitized path-str.
    """
    home = Path.home()
    if path.is_relative_to(home):
        return f"$HOME{os.path.sep}{path.relative_to(home)}"

    return f"{path}"


def get_package_path(package_name: str) -> Path:
    """
    Get the path to a package from site-packages.

    Args:
        package_name (str): The name of the package.

    Returns:
        Path
    """
    try:
        dist = distribution(package_name)
    except PackageNotFoundError as err:
        raise ValueError(f"Package '{package_name}' not found in site-packages.") from err

    package_path = Path(str(dist.locate_file(""))) / package_name
    if not package_path.exists():
        raise ValueError(f"Package '{package_name}' not found in site-packages.")

    return package_path


def extract_archive(archive_file: Path, destination: Optional[Path] = None):
    """
    Extract an archive file. Supports ``.zip`` or ``.tar.gz``.

    Args:
        archive_file (Path): The file-path to the archive.
        destination (Optional[Path]): Optionally provide a destination.
          Defaults to the parent directory of the archive file.
    """
    destination = destination or archive_file.parent
    if archive_file.suffix == ".zip":
        with zipfile.ZipFile(archive_file, "r") as zip_ref:
            zip_members = zip_ref.namelist()
            if top_level_dir := os.path.commonpath(zip_members):
                for zip_member in zip_members:
                    # Modify the member name to remove the top-level directory.
                    member_path = Path(zip_member)
                    relative_path = (
                        member_path.relative_to(top_level_dir) if top_level_dir else member_path
                    )
                    target_path = destination / relative_path

                    if member_path.is_dir():
                        target_path.mkdir(parents=True, exist_ok=True)
                    else:
                        target_path.parent.mkdir(parents=True, exist_ok=True)
                        with zip_ref.open(member_path.as_posix()) as source:
                            target_path.write_bytes(source.read())

            else:
                zip_ref.extractall(f"{destination}")

    elif archive_file.name.endswith(".tar.gz"):
        with tarfile.open(archive_file, "r:gz") as tar_ref:
            tar_members = tar_ref.getmembers()
            if top_level_dir := os.path.commonpath([m.name for m in tar_members]):
                for tar_member in tar_members:
                    # Modify the member name to remove the top-level directory.
                    tar_member.name = os.path.relpath(tar_member.name, top_level_dir)
                    tar_ref.extract(tar_member, path=destination)

            else:
                tar_ref.extractall(path=f"{destination}")

    else:
        raise ValueError(f"Unsupported zip format: '{archive_file.suffix}'.")


def _remove_readonly(func, path, excinfo):
    """
    Error handler for shutil.rmtree that handles removing read-only files.
    """
    os.chmod(path, stat.S_IWRITE)
    func(path)


class CacheDirectory:
    """
    A directory for caching data where each data item is named
    ``<key>.json`` and is in the directory. You can access the
    items by their key like a dictionary. This type is used
    in Ape's contract-caching for ContractTypes, ProxyInfoAPI,
    and other model types.
    """

    def __init__(self, path: Path):
        if path.is_file():
            raise ValueError("Expecting directory.")

        self._path = path

    def __getitem__(self, key: str) -> dict:
        """
        Get the data from ``base_path / <key>.json``.

        Returns:
            The JSON dictionary
        """
        return self.get_data(key)

    def __setitem__(self, key: str, value: dict):
        """
        Cache the given data to ``base_path / <key>.json``.

        Args:
            key (str): The key, used as the file name ``{key}.json``.
            value (dict): The JSON dictionary to cache.
        """
        self.cache_data(key, value)

    def __delitem__(self, key: str):
        """
        Delete the cache-file.

        Args:
            key (str): The file stem of the JSON.
        """
        self.delete_data(key)

    def get_file(self, key: str) -> Path:
        return self._path / f"{key}.json"

    def cache_data(self, key: str, data: dict):
        json_str = json.dumps(data)
        file = self.get_file(key)
        file.unlink(missing_ok=True)
        file.parent.mkdir(parents=True, exist_ok=True)
        file.write_text(json_str)

    def get_data(self, key: str) -> dict:
        file = self.get_file(key)
        if not file.is_file():
            return {}

        json_str = file.read_text(encoding="utf8")
        return json.loads(json_str)

    def delete_data(self, key: str):
        file = self.get_file(key)
        file.unlink(missing_ok=True)


@contextmanager
def within_directory(directory: Path):
    """
    A context-manager for changing the cwd to the given path.

    Args:
        directory (Path): The directory to change.
    """
    here = Path.cwd()
    if directory != here:
        os.chdir(directory)
    try:
        yield
    finally:
        if Path.cwd() != here:
            os.chdir(here)


class ChangeDirectory:
    """
    A context-manager for changing a directory. Initializing it
    will still change the directory, but you can optionally exit
    out of it to restore back to the original directory. Additionally,
    provides hooks to run when each of these events occur.
    """

    def __init__(
        self,
        original_path: Path,
        new_path: Path,
        chdir: Optional[Callable[[Path], None]] = None,
        on_push: Optional[Callable[[Path], dict]] = None,
        on_pop: Optional[Callable[[dict], None]] = None,
    ):
        self.original_path = original_path
        self.new_path = new_path
        self._on_push = on_push
        self._on_pop = on_pop
        self._chdir = chdir or os.chdir

        # Initiate the change now so you can still use this class
        # on methods that are not intended to be used in a context.
        if self.original_path != self.new_path:
            self._chdir(new_path)
            self._cache: dict = {} if self._on_push is None else self._on_push(new_path)
            self._did_change = True
        else:
            self._cache = {}
            self._did_change = False

    def __enter__(self):
        return self.new_path

    def __exit__(self, *args):
        if not self._did_change:
            # Don't do anything. Nothing changed.
            return

        # Handle the return to the original path.
        self._chdir(self.original_path)
        if self._on_pop:
            self._on_pop(self._cache)



import json
import re
from typing import Optional

import pytest
from eth_account import Account
from eth_account.hdaccount import ETHEREUM_DEFAULT_PATH

from ape.logging import HIDDEN_MESSAGE
from tests.conftest import GETH_URI, geth_process_test
from tests.integration.cli.utils import assert_failure, run_once

ALIAS = "test"
PASSWORD = "asdf1234"
PRIVATE_KEY = "0000000000000000000000000000000000000000000000000000000000000001"
MNEMONIC = "test test test test test test test test test test test junk"
INVALID_MNEMONIC = "test test"
CUSTOM_HDPATH = "m/44'/61'/0'/0/0"  # Ethereum Classic ($ETC) HDPath


def extract_mnemonic(output: str) -> Optional[list[str]]:
    found = re.search(r"Newly generated mnemonic is: ([a-z ]+)", output)
    if found:
        try:
            mnemonic_string = found.group(1)
            return mnemonic_string.split(" ")
        except IndexError:
            pass
    return None


@pytest.fixture(autouse=True)
def temp_keyfile_path(config):
    # NOTE: Is a fresh account for each use.
    path = config.DATA_FOLDER / "accounts" / f"{ALIAS}.json"
    path.parent.mkdir(parents=True, exist_ok=True)  # create accounts subfolder
    path.unlink(missing_ok=True)
    return path


@pytest.fixture
def temp_keyfile(temp_keyfile_path, keyparams):
    temp_keyfile_path.write_text(json.dumps(keyparams), encoding="utf8")

    yield temp_keyfile_path

    if temp_keyfile_path.is_file():
        temp_keyfile_path.unlink()


@pytest.fixture
def temp_account():
    return Account.from_key(bytes.fromhex(PRIVATE_KEY))


@pytest.fixture
def temp_account_mnemonic_default_hdpath():
    Account.enable_unaudited_hdwallet_features()
    return Account.from_mnemonic(MNEMONIC, account_path=ETHEREUM_DEFAULT_PATH)


@pytest.fixture
def temp_account_mnemonic_custom_hdpath():
    Account.enable_unaudited_hdwallet_features()
    return Account.from_mnemonic(MNEMONIC, account_path=CUSTOM_HDPATH)


@run_once
def test_import_valid_private_key(ape_cli, runner, temp_account, temp_keyfile_path):
    assert not temp_keyfile_path.is_file()
    # Add account from valid private key
    result = runner.invoke(
        ape_cli,
        ("accounts", "import", ALIAS),
        input="\n".join([f"0x{PRIVATE_KEY}", PASSWORD, PASSWORD]),
    )
    assert result.exit_code == 0, result.output
    assert temp_account.address in result.output
    assert ALIAS in result.output
    assert temp_keyfile_path.is_file()


@run_once
def test_import_alias_is_private_key(ape_cli, runner):
    # Attempt using private key as the alias.
    key_alias = f"0x{PRIVATE_KEY}"
    result = runner.invoke(
        ape_cli,
        ("accounts", "import", key_alias),
        input="\n".join([f"0x{PRIVATE_KEY}", PASSWORD, PASSWORD]),
    )
    assert result.exit_code != 0, result.output
    assert "ERROR" in result.output
    expected = "(AccountsError) Longer aliases cannot be hex strings.\n"
    assert expected in result.output


@run_once
def test_import_alias_is_really_long(ape_cli, runner):
    """
    For entropy related use-cases regarding alias, we
    must ensure long aliases are supported.
    """

    long_alias = "this is a long alias that i am going to use and you can't stop me"
    result = runner.invoke(
        ape_cli,
        ("accounts", "import", long_alias),
        input="\n".join([f"0x{PRIVATE_KEY}", PASSWORD, PASSWORD]),
    )
    assert result.exit_code == 0


@run_once
def test_import_invalid_private_key(ape_cli, runner):
    # Add account from invalid private key
    result = runner.invoke(
        ape_cli,
        ("accounts", "import", ALIAS),
        input="\n".join(["0xhello", PASSWORD, PASSWORD]),
    )
    assert result.exit_code == 1, result.output
    assert_failure(result, "Key can't be imported: Non-hexadecimal digit found")


@run_once
def test_import_alias_already_in_use(ape_cli, runner):
    def invoke_import():
        return runner.invoke(
            ape_cli,
            ("accounts", "import", ALIAS),
            input="\n".join([f"0x{PRIVATE_KEY}", PASSWORD, PASSWORD]),
        )

    result = invoke_import()
    assert result.exit_code == 0, result.output
    result = invoke_import()
    assert_failure(result, f"Account with alias '{ALIAS}' already in use")


@run_once
def test_import_account_instantiation_failure(mocker, ape_cli, runner):
    eth_account_from_key_patch = mocker.patch("ape_accounts._cli._account_from_key")
    eth_account_from_key_patch.side_effect = Exception("Can't instantiate this account!")
    result = runner.invoke(
        ape_cli,
        ("accounts", "import", ALIAS),
        input="\n".join([f"0x{PRIVATE_KEY}", PASSWORD, PASSWORD]),
    )
    assert_failure(result, "Key can't be imported: Can't instantiate this account!")


@run_once
def test_import_mnemonic_default_hdpath(
    ape_cli, runner, temp_account_mnemonic_default_hdpath, temp_keyfile_path
):
    assert not temp_keyfile_path.is_file()
    # Add account from mnemonic with default hdpath of ETHEREUM_DEFAULT_PATH
    result = runner.invoke(
        ape_cli,
        ("accounts", "import", "--use-mnemonic", ALIAS),
        input="\n".join([MNEMONIC, PASSWORD, PASSWORD]),
    )
    assert result.exit_code == 0, result.output
    assert temp_account_mnemonic_default_hdpath.address in result.output
    assert ALIAS in result.output
    assert temp_keyfile_path.is_file()


@run_once
def test_import_mnemonic_custom_hdpath(
    ape_cli, runner, temp_account_mnemonic_custom_hdpath, temp_keyfile_path
):
    assert not temp_keyfile_path.is_file()
    # Add account from mnemonic with custom hdpath
    result = runner.invoke(
        ape_cli,
        ("accounts", "import", ALIAS, "--use-mnemonic", "--hd-path", CUSTOM_HDPATH),
        input="\n".join([MNEMONIC, PASSWORD, PASSWORD]),
    )
    assert result.exit_code == 0, result.output
    assert temp_account_mnemonic_custom_hdpath.address in result.output
    assert ALIAS in result.output
    assert temp_keyfile_path.is_file()


@run_once
def test_export(ape_cli, runner, temp_keyfile, keyfile_account, accounts):
    # export key
    result = runner.invoke(
        ape_cli,
        ("accounts", "export", ALIAS),
        input="\n".join([PASSWORD, PASSWORD]),
    )
    assert result.exit_code == 0, result.output
    # NOTE: temp_keyfile uses the same address as the keyfile account.
    assert keyfile_account.address in result.output
    # NOTE: Both of these accounts are the same as the first
    #   test account.
    assert accounts[0].private_key in result.output


@run_once
def test_import_invalid_mnemonic(ape_cli, runner):
    # Add account from invalid mnemonic
    result = runner.invoke(
        ape_cli,
        ("accounts", "import", "--use-mnemonic", ALIAS),
        input="\n".join([INVALID_MNEMONIC, PASSWORD, PASSWORD]),
    )
    assert result.exit_code == 1, result.output
    assert_failure(result, "Seed phrase can't be imported")
    assert HIDDEN_MESSAGE in result.output
    assert INVALID_MNEMONIC not in result.output


@run_once
def test_generate_default(ape_cli, runner, temp_keyfile_path):
    assert not temp_keyfile_path.is_file()
    # Generate new private key
    show_mnemonic = ""
    result = runner.invoke(
        ape_cli,
        ("accounts", "generate", ALIAS),
        input="\n".join(["random entropy", show_mnemonic, PASSWORD, PASSWORD]),
    )
    assert result.exit_code == 0, result.output
    mnemonic = extract_mnemonic(result.output)
    assert mnemonic is not None
    mnemonic_length = len(mnemonic)
    assert mnemonic_length == 12
    assert ETHEREUM_DEFAULT_PATH in result.output
    assert ALIAS in result.output
    assert temp_keyfile_path.is_file()


@run_once
def test_generate_hide_mnemonic_prompt(ape_cli, runner, temp_keyfile_path):
    assert not temp_keyfile_path.is_file()
    # Generate new private key
    show_mnemonic = "n"
    result = runner.invoke(
        ape_cli,
        ("accounts", "generate", ALIAS),
        input="\n".join(["random entropy", show_mnemonic, PASSWORD, PASSWORD]),
    )
    assert result.exit_code == 0, result.output
    assert "Newly generated mnemonic is" not in result.output
    assert ETHEREUM_DEFAULT_PATH in result.output
    assert ALIAS in result.output
    assert temp_keyfile_path.is_file()


@run_once
def test_generate_hide_mnemonic_option(ape_cli, runner, temp_keyfile_path):
    assert not temp_keyfile_path.is_file()
    # Generate new private key
    result = runner.invoke(
        ape_cli,
        ("accounts", "generate", ALIAS, "--hide-mnemonic"),
        input="\n".join(["random entropy", PASSWORD, PASSWORD]),
    )
    assert result.exit_code == 0, result.output
    assert "Newly generated mnemonic is" not in result.output
    assert ETHEREUM_DEFAULT_PATH in result.output
    assert ALIAS in result.output
    assert temp_keyfile_path.is_file()


@run_once
def test_generate_24_words(ape_cli, runner, temp_keyfile_path):
    assert not temp_keyfile_path.is_file()
    # Generate new private key
    show_mnemonic = ""
    word_count = 24
    result = runner.invoke(
        ape_cli,
        ("accounts", "generate", ALIAS, "--word-count", word_count),
        input="\n".join(["random entropy", show_mnemonic, PASSWORD, PASSWORD]),
    )
    assert result.exit_code == 0, result.output
    mnemonic = extract_mnemonic(result.output)
    assert mnemonic is not None
    mnemonic_length = len(mnemonic)
    assert mnemonic_length == word_count
    assert ETHEREUM_DEFAULT_PATH in result.output
    assert ALIAS in result.output
    assert temp_keyfile_path.is_file()


@run_once
def test_generate_custom_hdpath(ape_cli, runner, temp_keyfile_path):
    assert not temp_keyfile_path.is_file()
    # Generate new private key
    show_mnemonic = ""
    result = runner.invoke(
        ape_cli,
        ("accounts", "generate", ALIAS, "--hd-path", CUSTOM_HDPATH),
        input="\n".join(["random entropy", show_mnemonic, PASSWORD, PASSWORD]),
    )
    assert result.exit_code == 0, result.output
    mnemonic = extract_mnemonic(result.output)
    assert mnemonic is not None
    mnemonic_length = len(mnemonic)
    assert mnemonic_length == 12
    assert CUSTOM_HDPATH in result.output
    assert ALIAS in result.output
    assert temp_keyfile_path.is_file()


@run_once
def test_generate_24_words_and_custom_hdpath(ape_cli, runner, temp_keyfile_path):
    assert not temp_keyfile_path.is_file()
    # Generate new private key
    show_mnemonic = ""
    word_count = 24
    result = runner.invoke(
        ape_cli,
        ("accounts", "generate", ALIAS, "--word-count", word_count, "--hd-path", CUSTOM_HDPATH),
        input="\n".join(["random entropy", show_mnemonic, PASSWORD, PASSWORD]),
    )
    assert result.exit_code == 0, result.output
    mnemonic = extract_mnemonic(result.output)
    assert mnemonic is not None
    mnemonic_length = len(mnemonic)
    assert mnemonic_length == word_count
    assert CUSTOM_HDPATH in result.output
    assert ALIAS in result.output
    assert temp_keyfile_path.is_file()


@run_once
def test_generate_alias_already_in_use(ape_cli, runner):
    def invoke_generate():
        show_mnemonic = ""
        return runner.invoke(
            ape_cli,
            ("accounts", "generate", ALIAS),
            input="\n".join(["random entropy", show_mnemonic, PASSWORD, PASSWORD]),
        )

    result = invoke_generate()
    assert result.exit_code == 0, result.output
    result = invoke_generate()
    assert_failure(result, f"Account with alias '{ALIAS}' already in use")


@run_once
def test_list(ape_cli, runner, keyfile_account):
    result = runner.invoke(ape_cli, ("accounts", "list"), catch_exceptions=False)
    assert keyfile_account.alias in result.output
    assert keyfile_account.address in result.output


@run_once
def test_list_all(ape_cli, runner, keyfile_account):
    result = runner.invoke(ape_cli, ("accounts", "list", "--all"), catch_exceptions=False)
    assert keyfile_account.alias in result.output
    assert keyfile_account.address in result.output


@run_once
def test_change_password(ape_cli, runner, temp_keyfile):
    assert temp_keyfile.is_file()
    # Delete Account (`N` for "Leave unlocked?")
    valid_input = [PASSWORD, "N", "password2", "password2"]
    result = runner.invoke(
        ape_cli,
        ("accounts", "change-password", ALIAS),
        input="\n".join(valid_input) + "\n",
    )
    assert result.exit_code == 0, result.output


@run_once
def test_delete(ape_cli, runner, temp_keyfile):
    assert temp_keyfile.is_file()
    # Delete Account
    result = runner.invoke(ape_cli, ("accounts", "delete", ALIAS), input=f"{PASSWORD}\n")
    assert result.exit_code == 0, result.output
    assert not temp_keyfile.is_file()


@run_once
@geth_process_test
@pytest.mark.xfail(reason="https://github.com/ethereum/go-ethereum/issues/31967")
def test_authorizations_cli(ape_cli, runner, keyfile_account, geth_contract):
    result = runner.invoke(
        ape_cli,
        (
            "accounts",
            "auth",
            "show",
            "--account",
            keyfile_account.alias,
            "--network",
            GETH_URI,
        ),
        catch_exceptions=False,
    )
    assert f"{keyfile_account.address} has no delegate" in result.output

    result = runner.invoke(
        ape_cli,
        (
            "accounts",
            "auth",
            "set",
            geth_contract.address,
            # NOTE: 0x23fd0e40 is method_id for `myNumber()`
            # NOTE: Must call something, since `__default__` raises
            "--data",
            "0x23fd0e40",
            "--account",
            keyfile_account.alias,
            "--network",
            GETH_URI,
        ),
        catch_exceptions=False,
        input="\n".join(["y", PASSWORD, "y", "y"]),
    )
    assert result.exit_code == 0, result.output

    result = runner.invoke(
        ape_cli,
        (
            "accounts",
            "auth",
            "show",
            "--account",
            keyfile_account.alias,
            "--network",
            GETH_URI,
        ),
        catch_exceptions=False,
    )
    assert f"{keyfile_account.address} is delegated to {geth_contract.address}" in result.output

    result = runner.invoke(
        ape_cli,
        (
            "accounts",
            "auth",
            "rm",
            "--account",
            keyfile_account.alias,
            "--network",
            GETH_URI,
        ),
        catch_exceptions=False,
        input="\n".join(["y", PASSWORD, "y", "y"]),
    )
    assert result.exit_code == 0, result.output

    result = runner.invoke(
        ape_cli,
        (
            "accounts",
            "auth",
            "show",
            "--account",
            keyfile_account.alias,
            "--network",
            GETH_URI,
        ),
        catch_exceptions=False,
    )
    assert f"{keyfile_account.address} has no delegate" in result.output



import sys
from importlib import import_module
from pathlib import Path
from typing import TYPE_CHECKING, Optional

import click

from ape.cli.options import ape_cli_context, config_override_option
from ape.exceptions import ProjectError
from ape.logging import logger

if TYPE_CHECKING:
    from ape.managers.project import Dependency


@click.group()
def cli():
    """
    Package management tools
    """


@cli.command("list")
@ape_cli_context()
@click.option("--all", "list_all", help="List all installed dependencies", is_flag=True)
def _list(cli_ctx, list_all):
    """
    List installed packages
    """
    dm = cli_ctx.dependency_manager
    packages = []
    dependencies = [*list(dm.get_project_dependencies(use_cache=True, allow_install=False))]
    if list_all:
        dependencies = list({*dependencies, *dm.installed})

    for dependency in dependencies:
        if dependency.installed:
            is_installed = True
            try:
                is_compiled = dependency.project.is_compiled
            except ProjectError:
                # Project may not even be installed right.
                is_compiled = False

        else:
            is_installed = False
            is_compiled = False

        # For local dependencies, use the short name.
        # This is mostly because it looks nicer than very long paths.
        dependency_module = import_module("ape_pm.dependency")
        name = (
            dependency.name
            if isinstance(dependency.api, dependency_module.LocalDependency)
            else dependency.package_id
        )

        try:
            version = dependency.version
        except ProjectError:
            version = "<error>"

        item = {
            "name": name,
            "version": version,
            "installed": is_installed,
            "compiled": is_compiled,
        }
        packages.append(item)

    if not packages:
        if list_all:
            click.echo("No packages installed.")
        else:
            click.echo("No packages installed. Use `--all` to list all installed dependencies.")
        return

    # Output gathered packages.
    longest_name = max([4, *[len(p["name"]) for p in packages]])
    longest_version = max([7, *[len(p["version"]) for p in packages]])
    longest_installed = max([9, *[len(f"{p['installed']}") for p in packages]])
    tab = "  "

    header_name_space = ((longest_name - len("NAME")) + 2) * " "
    version_name_space = ((longest_version - len("VERSION")) + 2) * " "

    def get_package_str(_package) -> str:
        dep_name = click.style(_package["name"], bold=True)
        version = _package["version"]
        installed = (
            click.style(_package["installed"], fg="green") if _package.get("installed") else "False"
        )
        compiled = (
            click.style(_package["compiled"], fg="green") if _package.get("compiled") else "False"
        )
        spacing_name = ((longest_name - len(_package["name"])) + len(tab)) * " "
        spacing_version = ((longest_version - len(version)) + len(tab)) * " "
        spacing_installed = ((longest_installed - len(f"{_package['installed']}")) + len(tab)) * " "
        return (
            f"{dep_name}{spacing_name}{version}{spacing_version}"
            f"{installed}{spacing_installed}{compiled}"
        )

    def rows():
        yield f"NAME{header_name_space}VERSION{version_name_space}INSTALLED  COMPILED\n"
        for _package in sorted(packages, key=lambda p: f"{p['name']}{p['version']}"):
            yield f"{get_package_str(_package)}\n"

    if len(packages) > 16:
        click.echo_via_pager(rows())
    else:
        for row in rows():
            click.echo(row.strip())


def _handle_package_path(path: Path, original_value: Optional[str] = None) -> dict:
    if not path.exists():
        value = original_value or path.as_posix()
        raise click.BadArgumentUsage(f"Unknown package '{value}'.")

    elif path.is_file() and (path.stem == "ape-config" or path.name == "pyproject.toml"):
        path = path.parent

    path = path.resolve().absolute()
    return {"local": path.as_posix()}


def _package_callback(ctx, param, value):
    if value is None:
        # Install all packages from local project.
        return None

    elif isinstance(value, Path):
        return _handle_package_path(value)

    elif value.startswith("gh:"):
        # Is a GitHub style dependency
        return {"github": value[3:]}

    elif value.startswith("npm:"):
        # Is an NPM style dependency
        return {"npm": value[4:]}

    elif value == ".":
        return value

    # Check if is a local package.
    try:
        path = Path(value).absolute()
    except Exception:
        pass
    else:
        if path.exists():
            return _handle_package_path(path, original_value=value)

    if isinstance(value, str) and ":" in value:
        # Catch-all for unknown dependency types that may exist.
        parts = value.split(":")
        return {parts[0]: parts[1]}

    elif isinstance(value, str):
        from ape import project

        version = ctx.params.get("version")
        name = value
        if not version and "@" in value:
            name, version = value.split("@")

        try:
            dependency = project.dependencies.get_dependency_api(name, version=version)
        except ProjectError:
            pass
        else:
            return dependency

    raise click.BadArgumentUsage(f"Unknown package '{value}'.")


@cli.command()
@ape_cli_context()
@click.argument("package", required=False, callback=_package_callback)
@click.option("--name", help="The name of the dependency", metavar="NAME")
@click.option("--version", help="The dependency's version", metavar="VERSION")
@click.option(
    "--ref",
    help="A reference flag, used for GitHub branches or tags instead of version",
    metavar="REF",
)
@click.option("--force", "-f", help="Force a re-install", is_flag=True)
@config_override_option()
@click.option("--no-recurse", is_flag=True, help="Avoids installing dependencies of dependencies")
def install(cli_ctx, package, name, version, ref, force, config_override, no_recurse):
    """
    Download and cache packages
    """

    pm = cli_ctx.local_project
    if not package or package == ".":
        if version:
            cli_ctx.abort("Cannot specify version when installing from config.")

        pm.dependencies.install(use_cache=not force, recurse=not no_recurse)
        message = "All project packages installed."

        # In the case the user didn't realize --force is required to re-install.
        if not force:
            message = f"{message} Use `--force` to re-install."

        cli_ctx.logger.success(message)
        return

    from ape.api.projects import DependencyAPI

    if isinstance(package, DependencyAPI):
        package = package.model_dump()

    if name:
        package["name"] = name
    if ref:
        package["ref"] = ref
    if version:
        package["version"] = version
    if config_override:
        package["config_override"] = config_override

    try:
        dependency = pm.dependencies.install(**package, use_cache=not force, recurse=not no_recurse)
    except Exception as err:
        cli_ctx.logger.log_error(err)
        sys.exit(1)

    cli_ctx.logger.success(f"Package '{dependency.name}@{dependency.version}' installed.")


@cli.command()
@ape_cli_context()
@click.argument("name", required=False)
@click.argument("versions", nargs=-1, required=False)
@click.option(
    "-y", "--yes", is_flag=True, help="Automatically confirm the removal of the package(s)"
)
def uninstall(cli_ctx, name, versions, yes):
    """
    Uninstall a package

    This command removes a package from the installed packages.

    If specific versions are provided, only those versions of the package will be
    removed. If no versions are provided, the command will prompt you to choose
    versions to remove. You can also choose to remove all versions of the package.

    Examples:\n
    - Remove specific versions: ape pm uninstall <PackageName> "1.0.0" "2.0.0"\n
    - Prompt to choose versions: ape pm uninstall <PackageName>\n
    - Remove all versions: ape pm uninstall <PackageName> -y
    """

    pm = cli_ctx.local_project

    # NOTE: Purposely don't call `get_dependency` or anything so we for sure
    #   are only checking the installed.
    installed = {d for d in pm.dependencies.installed}

    did_error = False
    did_find = False

    if not name or name == ".":
        if versions:
            cli_ctx.abort("Cannot specify version when uninstalling from config.")

        # Uninstall all dependencies from the config.
        for cfg in pm.config.dependencies:
            api = pm.dependencies.decode_dependency(**cfg)
            for dependency in installed:
                if dependency.name != api.name or dependency.version != api.version_id:
                    continue

                did_find = True
                res = _uninstall(dependency, yes=yes)
                if res is False:
                    did_error = True

    else:
        deps_to_remove = {
            d
            for d in installed
            if (d.name == name or d.package_id == name)
            and (d.version in versions if versions else True)
        }
        for dependency in deps_to_remove:
            did_find = True
            res = _uninstall(dependency, yes=yes)
            if res is False:
                did_error = True

    if not did_find:
        if name:
            name = ", ".join([f"{name}={v}" for v in versions]) if versions else name
            cli_ctx.logger.error(f"Package(s) '{name}' not installed.")
        else:
            cli_ctx.logger.error(
                "No package(s) installed in local project. "
                "Please specify a package to uninstall or go to a local project."
            )

        did_error = True

    sys.exit(int(did_error))


def _uninstall(dependency: "Dependency", yes: bool = False) -> bool:
    key = f"{dependency.name}={dependency.version}"
    if not yes and not click.confirm(f"Remove '{key}'"):
        return True  # Not an error.

    try:
        dependency.uninstall()
    except Exception as err:
        logger.error(f"Failed uninstalling '{key}': {err}")
        return False

    logger.success(f"Uninstalled '{key}'.")
    return True


@cli.command()
@ape_cli_context()
@click.argument("name", required=False)
@click.option("--version", help="The dependency version", metavar="VERSION")
@click.option("--force", "-f", help="Force a re-compile", is_flag=True)
@config_override_option()
def compile(cli_ctx, name, version, force, config_override):
    """
    Compile a package
    """
    pm = cli_ctx.local_project
    if not name or name == ".":
        if version:
            cli_ctx.abort("Cannot specify 'version' without 'name'.")

        # Compile all from config.
        did_error = False
        for cfg in pm.config.dependencies:
            if config_override:
                cfg["config_override"] = config_override

            dependency = pm.dependencies.install(**cfg)
            _compile_dependency(cli_ctx, dependency, force)

        if did_error:
            sys.exit(1)

        return

    if version:
        to_compile = [pm.dependencies.get_dependency(name, version)]
    else:
        to_compile = [d for d in pm.dependencies.get_versions(name)]

    if not to_compile:
        key = f"{name}@{version}" if version else name
        cli_ctx.abort(f"Dependency '{key}' unknown. Is it installed?")

    for dependency in to_compile:
        if config_override:
            dependency.api.config_override = config_override

        _compile_dependency(cli_ctx, dependency, force)


def _compile_dependency(cli_ctx, dependency: "Dependency", force: bool):
    try:
        result = dependency.compile(use_cache=not force)
    except Exception as err:
        cli_ctx.logger.error(str(err))
    else:
        if result:
            cli_ctx.logger.success(f"Package '{dependency.name}@{dependency.version}' compiled.")
        # else: user should have received warning from `dependency.compile()` if there
        # was no result.



import inspect
import sys
from collections.abc import Callable
from functools import partial
from pathlib import Path
from typing import TYPE_CHECKING, Any, NoReturn, Optional, Union

import click
from click import Option

from ape.cli.choices import (
    _ACCOUNT_TYPE_FILTER,
    _NONE_NETWORK,
    AccountAliasPromptChoice,
    NetworkChoice,
    OutputFormat,
    output_format_choice,
)
from ape.cli.commands import ConnectedProviderCommand
from ape.cli.paramtype import JSON, Noop
from ape.exceptions import Abort, ProjectError
from ape.logging import DEFAULT_LOG_LEVEL, ApeLogger, LogLevel, logger

if TYPE_CHECKING:
    from ethpm_types.contract_type import ContractType

_VERBOSITY_VALUES = ("--verbosity", "-v")


class ApeCliContextObject(dict):
    """
    A ``click`` context object class. Use via :meth:`~ape.cli.options.ape_cli_context()`.
    It provides common CLI utilities for ape, such as logging or
    access to the managers.
    """

    def __init__(self):
        self.logger = logger
        super().__init__({})

    def __repr__(self) -> str:
        # Customizing this because otherwise it uses `dict` repr, which is confusing.
        return f"<{self.__class__.__name__}>"

    def __getattr__(self, item: str) -> Any:
        try:
            return self.__getattribute__(item)
        except AttributeError:
            from ape.utils.basemodel import ManagerAccessMixin

            return getattr(ManagerAccessMixin, item)

    @staticmethod
    def abort(msg: str, base_error: Optional[Exception] = None) -> NoReturn:
        """
        End execution of the current command invocation.

        Args:
            msg (str): A message to output to the terminal.
            base_error (Exception, optional): Optionally provide
              an error to preserve the exception stack.
        """

        if base_error:
            logger.error(msg)
            raise Abort(msg) from base_error

        raise Abort(msg)


def verbosity_option(
    cli_logger: Optional[ApeLogger] = None,
    default: Optional[Union[str, int, LogLevel]] = None,
    callback: Optional[Callable] = None,
    **kwargs,
) -> Callable:
    """A decorator that adds a `--verbosity, -v` option to the decorated
    command.

    Args:
        cli_logger (:class:`~ape.logging.ApeLogger` | None): Optionally pass
          a custom logger object.
        default (str | int | :class:`~ape.logging.LogLevel`): The default log-level
          for this command.
        callback (Callable | None): A callback handler for passed-in verbosity values.
        **kwargs: Additional click overrides.

    Returns:
        click option
    """
    _logger = cli_logger or logger
    default = logger.level if default is None else default
    kwarguments = _create_verbosity_kwargs(
        _logger=_logger, default=default, callback=callback, **kwargs
    )
    return lambda f: click.option(*_VERBOSITY_VALUES, **kwarguments)(f)


def _create_verbosity_kwargs(
    _logger: Optional[ApeLogger] = None,
    default: Optional[Union[str, int, LogLevel]] = None,
    callback: Optional[Callable] = None,
    **kwargs,
) -> dict:
    default = logger.level if default is None else default
    ape_logger = _logger or logger

    def set_level(ctx, param, value):
        if isinstance(value, str):
            value = value.upper()
            if value.startswith("LOGLEVEL."):
                value = value.split(".")[-1].strip()
            elif value in ("DISABLE", "NONE"):
                value = ape_logger.DISABLE_LEVEL

        if callback is not None:
            value = callback(ctx, param, value)

        if ape_logger._did_parse_sys_argv:
            # Changing mid-session somehow (tests?)
            ape_logger.set_level(value)
        else:
            ape_logger._load_from_sys_argv(default=value)

    level_names = [lvl.name for lvl in LogLevel]
    names_str = f"{', '.join(level_names[:-1])}, or {level_names[-1]}"
    return {
        "callback": set_level,
        "default": default or DEFAULT_LOG_LEVEL,
        "metavar": "LVL",
        "expose_value": False,
        "help": f"One of {names_str}",
        "is_eager": True,
        "type": Noop(),
        **kwargs,
    }


def ape_cli_context(
    default_log_level: Optional[Union[str, int, LogLevel]] = None,
    obj_type: type = ApeCliContextObject,
) -> Callable:
    """
    A ``click`` context object with helpful utilities.
    Use in your commands to get access to common utility features,
    such as logging or accessing managers.

    Args:
        default_log_level (str | int | :class:`~ape.logging.LogLevel` |  None): The log-level
          value to pass to :meth:`~ape.cli.options.verbosity_option`.
        obj_type (Type): The context object type. Defaults to
          :class:`~ape.cli.options.ApeCliContextObject`. Sub-class
          the context to extend its functionality in your CLIs,
          such as if you want to add additional manager classes
          to the context.

    Returns:
        click option
    """
    default_log_level = logger.level if default_log_level is None else default_log_level

    def decorator(f):
        f = verbosity_option(logger, default=default_log_level)(f)
        f = click.make_pass_decorator(obj_type, ensure=True)(f)
        return f

    return decorator


class NetworkOption(Option):
    """
    The class used in `:meth:~ape.cli.options.network_option`.
    """

    # NOTE: Has to be kwargs only to avoid multiple-values for arg error.
    def __init__(self, *args, **kwargs) -> None:
        ecosystem = kwargs.pop("ecosystem", None)
        network = kwargs.pop("network", None)
        provider = kwargs.pop("provider", None)
        default = kwargs.pop("default", "auto")

        callback = kwargs.pop("callback", None)

        # NOTE: If using network_option, this part is skipped
        #  because parsing happens earlier to handle advanced usage.
        if not kwargs.get("type"):
            base_type = kwargs.pop("base_type", None)
            kwargs["type"] = NetworkChoice(
                case_sensitive=False,
                ecosystem=ecosystem,
                network=network,
                provider=provider,
                base_type=base_type,
                callback=callback,
            )
        elif callback is not None:
            # Make sure these are the same.
            kwargs["type"].callback = callback

        auto = default == "auto"
        required = kwargs.get("required", False)

        if auto and not required:
            if ecosystem:
                default = ecosystem[0] if isinstance(ecosystem, (list, tuple)) else ecosystem

            else:
                # NOTE: Use a function as the default so it is calculated lazily
                def fn():
                    from ape.utils.basemodel import ManagerAccessMixin

                    return ManagerAccessMixin.network_manager.default_ecosystem.name

                default = fn

        elif auto:
            default = None

        help_msg = (
            "Override the default network and provider. (see `ape networks list` for options)"
        )
        kwargs = {
            "param_decls": ("--network",),
            "help": help_msg,
            "default": default,
            "required": required,
            **kwargs,
        }
        super().__init__(**kwargs)


def network_option(
    default: Optional[Union[str, Callable]] = "auto",
    ecosystem: Optional[Union[list[str], str]] = None,
    network: Optional[Union[list[str], str]] = None,
    provider: Optional[Union[list[str], str]] = None,
    required: bool = False,
    **kwargs,
) -> Callable:
    """
    A ``click.option`` for specifying a network.

    Args:
        default (Optional[str]): Optionally, change which network to
          use as the default. Defaults to how ``ape`` normally
          selects a default network unless ``required=True``, then defaults to ``None``.
        ecosystem (Optional[Union[list[str], str]]): Filter the options by ecosystem.
          Defaults to getting all ecosystems.
        network (Optional[Union[list[str], str]]): Filter the options by network.
          Defaults to getting all networks in ecosystems.
        provider (Optional[Union[list[str], str]]): Filter the options by provider.
          Defaults to getting all providers in networks.
        required (bool): Whether the option is required. Defaults to ``False``.
          When set to ``True``, the default value is ``None``.
        kwargs: Additional overrides to ``click.option``.
    """

    def decorator(f):
        # These are the available network object names you can request.
        network_object_names = ("ecosystem", "network", "provider")
        requested_network_objects = _get_requested_networks(f, network_object_names)

        # When using network_option, handle parsing now so we can pass to
        # callback outside of command context.
        user_callback = kwargs.pop("callback", None)

        def callback(ctx, param, value):
            keep_as_choice_str = param.type.base_type is str
            try:
                provider_obj = _get_provider(value, default, keep_as_choice_str)
            except Exception as err:
                raise click.BadOptionUsage("--network", str(err), ctx)

            if provider_obj:
                _update_context_with_network(ctx, provider_obj, requested_network_objects)

            elif keep_as_choice_str:
                # Add raw choice to object context.
                ctx.obj = ctx.obj or {}
                ctx.params = ctx.params or {}
                ctx.obj["network"] = value
                ctx.params["network"] = value

            # else: provider is None, meaning not connected intentionally.

            return value if user_callback is None else user_callback(ctx, param, value)

        wrapped_f = _wrap_network_function(network_object_names, requested_network_objects, f)

        # Use NetworkChoice option.
        kwargs["type"] = None

        # Set this to false to avoid click passing in a str value for network.
        # This happens with `kwargs["type"] = None` and we are already handling
        # `network` via the partial.
        kwargs["expose_value"] = False

        # The callback will set any requests values in the command.
        kwargs["callback"] = callback

        # Create the actual option.
        return click.option(
            default=default,
            ecosystem=ecosystem,
            network=network,
            provider=provider,
            required=required,
            cls=NetworkOption,
            **kwargs,
        )(wrapped_f)

    return decorator


def _get_requested_networks(function, network_object_names):
    command_signature = inspect.signature(function)
    command_kwargs = [x.name for x in command_signature.parameters.values()]

    # Any combination of ["ecosystem", "network", "provider"]
    return [x for x in command_kwargs if x in network_object_names]


def _update_context_with_network(ctx, provider, requested_network_objects):
    choice_classes = {
        "ecosystem": provider.network.ecosystem,
        "network": provider.network,
        "provider": provider,
    }

    # Set the actual values in the callback.
    for item in requested_network_objects:
        instance = choice_classes[item]
        ctx.params[item] = instance

    if isinstance(ctx.command, ConnectedProviderCommand):
        # Place all values, regardless of request in
        # the context. This helps the Ape CLI backend.
        if ctx.obj is None:
            # Happens when using commands that don't use the
            # Ape context or any context.
            ctx.obj = {}

        for choice, obj in choice_classes.items():
            try:
                ctx.obj[choice] = obj
            except Exception:
                # This would only happen if using an unusual context object.
                raise Abort(
                    "Cannot use connected-provider command type(s) "
                    "with non key-settable context object."
                )


def _get_provider(value, default, keep_as_choice_str):
    from ape.api.providers import ProviderAPI
    from ape.utils.basemodel import ManagerAccessMixin

    use_default = value is None and default == "auto"

    if not keep_as_choice_str and use_default:
        default_ecosystem = ManagerAccessMixin.network_manager.default_ecosystem
        return default_ecosystem.default_network.default_provider

    elif value is None or keep_as_choice_str:
        return None

    elif isinstance(value, ProviderAPI):
        return value

    elif value == _NONE_NETWORK:
        return None

    else:
        network_ctx = ManagerAccessMixin.network_manager.parse_network_choice(value)
        return network_ctx._provider


def _wrap_network_function(network_object_names, requested_network_objects, function):
    # Prevent argument errors but initializing callback to use None placeholders.
    partial_kwargs: dict = {}
    for arg_type in network_object_names:
        if arg_type in requested_network_objects:
            partial_kwargs[arg_type] = None

    if partial_kwargs:
        wrapped_f = partial(function, **partial_kwargs)

        # NOTE: The following is needed for click internals.
        wrapped_f.__name__ = function.__name__  # type: ignore[attr-defined]

        # NOTE: The following is needed for sphinx internals.
        wrapped_f.__doc__ = function.__doc__

        # Add other click parameters.
        if hasattr(function, "__click_params__"):
            wrapped_f.__click_params__ = function.__click_params__  # type: ignore[attr-defined]

        return wrapped_f

    else:
        # No network kwargs are used. No need for partial wrapper.
        return function


def skip_confirmation_option(help="") -> Callable:
    """
    A ``click.option`` for skipping confirmation (``--yes``).

    Args:
        help (str): CLI option help text. Defaults to ``""``.
    """

    return click.option(
        "-y",
        "--yes",
        "skip_confirmation",
        default=False,
        is_flag=True,
        help=help,
    )


def _account_callback(ctx, param, value):
    if param and not value:
        return param.type.select_account()

    return value


def account_option(account_type: _ACCOUNT_TYPE_FILTER = None) -> Callable:
    """
    A CLI option that accepts either the account alias or the account number.
    If not given anything, it will prompt the user to select an account.
    """

    return click.option(
        "--account",
        type=AccountAliasPromptChoice(key=account_type),
        callback=_account_callback,
    )


def _load_contracts(ctx, param, value) -> Optional[Union["ContractType", list["ContractType"]]]:
    if not value:
        return None

    from ape.utils.basemodel import ManagerAccessMixin

    if len(ManagerAccessMixin.local_project.contracts) == 0:
        raise ProjectError("Project has no contracts.")

    # If the user passed in `multiple=True`, then `value` is a list,
    # and therefore we should also return a list.
    is_multiple = isinstance(value, (tuple, list))

    def get_contract(contract_name: str) -> "ContractType":
        if contract_name not in ManagerAccessMixin.local_project.contracts:
            raise ProjectError(f"No contract named '{value}'")

        return ManagerAccessMixin.local_project.contracts[contract_name]

    return [get_contract(c) for c in value] if is_multiple else get_contract(value)


def contract_option(help=None, required=False, multiple=False) -> Callable:
    """
    Contract(s) from the current project.
    If you pass ``multiple=True``, you will get a list of contract types from the callback.


        :class:`~ape.exceptions.ContractError`: In the callback when it fails to load the contracts.
    """

    help = help or "The name of a contract in the current project"
    return click.option(
        "--contract", help=help, required=required, callback=_load_contracts, multiple=multiple
    )


def output_format_option(default: OutputFormat = OutputFormat.TREE) -> Callable:
    """
    A ``click.option`` for specifying a format to use when outputting data.

    Args:
        default (:class:`~ape.cli.choices.OutputFormat`): Defaults to ``TREE`` format.
    """

    return click.option(
        "--format",
        "output_format",
        type=output_format_choice(),
        default=default.value,
        callback=lambda ctx, param, value: OutputFormat(value.upper()),
    )


def incompatible_with(incompatible_opts) -> type[click.Option]:
    """
    Factory for creating custom ``click.Option`` subclasses that
    enforce incompatibility with the option strings passed to this function.

    Usage example::

        import click

        @click.command()
        @click.option("--option", cls=incompatible_with(["other_option"]))
        def cmd(option, other_option):
            ....
    """

    if isinstance(incompatible_opts, str):
        incompatible_opts = [incompatible_opts]

    class IncompatibleOption(click.Option):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

        def handle_parse_result(self, ctx, opts, args):
            # if None it means we're in autocomplete mode and don't want to validate
            if ctx.obj is not None:
                found_incompatible = ", ".join(
                    [f"--{opt.replace('_', '-')}" for opt in opts if opt in incompatible_opts]
                )
                if self.name is not None and self.name in opts and found_incompatible:
                    name = self.name.replace("_", "-")
                    raise click.BadOptionUsage(
                        option_name=self.name,
                        message=f"'--{name}' can't be used with '{found_incompatible}'.",
                    )
            return super().handle_parse_result(ctx, opts, args)

    return IncompatibleOption


def _project_path_callback(ctx, param, val):
    return Path(val) if val else Path.cwd()


def _project_callback(ctx, param, val):
    if "--help" in sys.argv or "-h" in sys.argv:
        # Perf: project option is eager; have to check sys.argv to
        #  know to exit early when only doing --help.
        return

    from ape.utils.basemodel import ManagerAccessMixin

    pm = None
    if not val:
        pm = ManagerAccessMixin.local_project

    else:
        path = Path(val)
        if path == ManagerAccessMixin.local_project.path:
            pm = ManagerAccessMixin.local_project

        else:
            Project = ManagerAccessMixin.Project
            if path.is_file() and path.suffix == ".json":
                pm = Project.from_manifest(path)

            elif path.is_dir():
                pm = Project(path)

    if pm is None:
        raise click.BadOptionUsage("--project", "Not a valid project")

    return pm


def project_option(**kwargs):
    _type = kwargs.pop("type", None)
    callback = (
        _project_path_callback
        if (isinstance(_type, type) and issubclass(_type, Path))
        else _project_callback
    )
    return click.option(
        "--project",
        help="The path to a local project or manifest",
        callback=callback,
        metavar="PATH",
        is_eager=True,
        **kwargs,
    )


def _json_option(name, help, **kwargs):
    return click.option(
        name,
        help=help,
        type=JSON(),
        metavar='{"KEY": "VAL"}',
        **kwargs,
    )


def config_override_option(**kwargs):
    return _json_option("--config-override", help="Config override mappings", **kwargs)


def _excluded_compilers_callback(ctx, param, value):
    if not value:
        return

    return [c.lower() for c in value]


def excluded_compilers_option(**kwargs):
    from ape.utils.basemodel import ManagerAccessMixin

    registered_compilers_options = [
        compiler.name
        for compiler in ManagerAccessMixin.compiler_manager.registered_compilers.values()
    ]

    return click.option(
        "--exclude-compiler",
        "excluded_compilers",
        help="Exclude specific compilers from the compilation process",
        type=click.Choice(registered_compilers_options, case_sensitive=False),
        callback=_excluded_compilers_callback,
        multiple=True,
        **kwargs,
    )
